#include "networkingandfiles.h"

NetworkingAndFiles::NetworkingAndFiles(bool isKiosk)
{
    this->isKiosk = isKiosk;

    errLogSize             = 0;
    logging                = isKiosk;
    mutex                  = nullptr;
    networkAccessManager   = nullptr;
    totalCachedBytes       = 0;
    uiLanguage             = "en";
    utilsProcess           = nullptr;
    wifiRetryTimestamp     = 0;

    if (isKiosk) {
        QDir logDir(QStandardPaths::writableLocation(QStandardPaths::AppDataLocation));
        errLog.setFileName(logDir.absoluteFilePath(".tef_cac.err"));

        QByteArray msg = QString("Error log at '%1'\n").arg(errLog.fileName()).toLatin1();
        std::cerr.write(msg.data(), msg.size());

        if (logDir.mkpath(logDir.absolutePath())) {
            if (errLog.exists()) {
                errLogSize = errLog.size();
            }
            if (!errLog.open(QIODevice::Append)) {
                logging = false;
                std::cerr << "Unable to open error log in read/write mode\n";
            }
        }
        else {
            logging = false;
            std::cerr << "Unable to create directory for error log\n";
        }
    }
}

NetworkingAndFiles::~NetworkingAndFiles()
{
    if (mutex != nullptr) {
        delete mutex;
        mutex = nullptr;
    }
    if (networkAccessManager != nullptr) {
        delete networkAccessManager;
        networkAccessManager = nullptr;
    }

    if (utilsProcess != nullptr) {
        if (utilsProcess->state() != QProcess::NotRunning) {
            utilsProcess->kill();
        }
        delete utilsProcess;
        utilsProcess = nullptr;
    }

    if (errLog.isOpen()) {
        errLog.close();
    }
}


void NetworkingAndFiles::checkConnection()
{
    execUtil("tef_inet_is_accessible");
}


void NetworkingAndFiles::execUtil(QString basename)
{
    execUtil(basename, QStringList());
}


void NetworkingAndFiles::execUtil(QString basename, QStringList arguments)
{
    if ((utilsProcess == nullptr) || (utilsProcess->state() != QProcess::NotRunning)) {
        return;
    }

    QSettings settings;

    QDir utilsDir(settings.value("ext/sysutils_dir", DEFAULT_SYSUTILS_DIR).toString());
    if (!utilsDir.exists()) {
        writeToErrorLog(QString("%1 does not exist").arg(utilsDir.absolutePath()));
        return;
    }

    QFileInfo utilsProg(utilsDir.absoluteFilePath(basename));
    if (!utilsProg.exists()) {
        writeToErrorLog(QString("%1 does not exist").arg(utilsProg.absolutePath()));
        return;
    }

    utilsProcess->start(utilsProg.canonicalFilePath(), arguments);
}


void NetworkingAndFiles::get(QUERIES query, int page, QVariantMap querySpecific)
{
    mutex->lock();
    QString lang = uiLanguage;
    mutex->unlock();

    if (repeatAfterConnected.url().isValid()) {
        writeToErrorLog(QString("NetworkingAndFiles::get was called while repeatAfterConnected is not null, query: %1").arg(query));
        return;
    }

    QUrl url;

    QUrlQuery queryString;
    if (page > 0) {
        queryString.addQueryItem("page", QString("%1").arg(page));
        queryString.addQueryItem("per_page", QString("%1").arg(QURAN_QUERY_PAGE_SIZE));
    }

    QObject *extra = new QObject();
    extra->setProperty("queryId", query);

    if (query == QUERY_COUNTRIES) {
        queryString.addQueryItem("username", geonamesUser);
        queryString.addQueryItem("lang", lang);
        url = QUrl("http://api.geonames.org/countryInfo");
    }
    else if (query == QUERY_LOCATION) {
        queryString.addQueryItem("access_key", ipStackAccessKey);
        url = QUrl("https://api.ipstack.com/check");
    }
    else if (query == QUERY_REGIONS) {
        queryString.addQueryItem("geonameId", querySpecific.value("geonameId").toString());
        queryString.addQueryItem("username", geonamesUser);
        queryString.addQueryItem("lang", lang);
        queryString.addQueryItem("style", "FULL");
        url = QUrl("http://api.geonames.org/children");
    }
    else if (query == LOCATION_SEARCH_POSTALCODE) {
        queryString.addQueryItem("country", querySpecific.value("country").toString());
        queryString.addQueryItem("postalcode", querySpecific.value("postalcode").toString());
        queryString.addQueryItem("username", geonamesUser);
        queryString.addQueryItem("lang", lang);
        queryString.addQueryItem("maxRows", QString("%1").arg(SEARCH_LIMIT));
        queryString.addQueryItem("style", "FULL");
        url = QUrl("http://api.geonames.org/postalCodeSearch");
    }
    else if (query == LOCATION_SEARCH_FREETEXT) {
        queryString.addQueryItem("country", querySpecific.value("country").toString());
        queryString.addQueryItem("name_startsWith", querySpecific.value("name_startsWith").toString());
        queryString.addQueryItem("username", geonamesUser);
        queryString.addQueryItem("lang", lang);
        queryString.addQueryItem("maxRows", QString("%1").arg(SEARCH_LIMIT));
        queryString.addQueryItem("style", "FULL");

        if (querySpecific.contains("adminCode1")) {
            queryString.addQueryItem("adminCode1", querySpecific.value("adminCode1").toString());
        }

        url = QUrl("http://api.geonames.org/search");
    }
    else if (query == LOCATION_TIMEZONE) {
        queryString.addQueryItem("lat", querySpecific.value("lat").toString());
        queryString.addQueryItem("lng", querySpecific.value("lng").toString());
        queryString.addQueryItem("username", geonamesUser);
        queryString.addQueryItem("lang", lang);
        queryString.addQueryItem("maxRows", QString("%1").arg(SEARCH_LIMIT));
        queryString.addQueryItem("style", "FULL");


        extra->setProperty("latitude",    querySpecific.value("lat"));
        extra->setProperty("longitude",   querySpecific.value("lng"));
        extra->setProperty("country",     querySpecific.value("country"));
        extra->setProperty("countryCode", querySpecific.value("countryCode"));
        extra->setProperty("region",      querySpecific.value("region"));
        extra->setProperty("city",        querySpecific.value("city"));

        url = QUrl("http://api.geonames.org/timezone");
    }
    else if (query == QURAN_SUWAR) {
        QVariantList cached = quranCacheLoad(quranCacheFileName(query, lang, page, querySpecific));
        if (!cached.isEmpty()) {
            emit got(query, true, cached);
            return;
        }

        queryString.addQueryItem("language", lang);
        url = QUrl("https://api.quran.com/api/v4/chapters");
    }
    else if (query == QURAN_AYAT) {
        if (!querySpecific.contains("surah")) {
            writeToErrorLog("Surah number is unknown");
            emit uiMsg("Surah number is unknown", true);
            emit got(query, false, 0);
            return;
        }

        QVariantMap loadExtra(querySpecific);
        loadExtra.remove("ayah");
        QVariantList cached = quranCacheLoad(quranCacheFileName(query, lang, page, loadExtra));
        if (!cached.isEmpty()) {
            emit got(query, true, QVariantMap({ {"surah", querySpecific.value("surah")}, { "ayat", cached }, { "selectedAyah", querySpecific.value("surah") } }));
            return;
        }

        queryString.addQueryItem("audio", "2");
        queryString.addQueryItem("words", "true");
        queryString.addQueryItem("word_fields", "id,position,page_number,code_v2");
        queryString.addQueryItem("fields", "id,verse_number,text_uthmani_simple");
        queryString.addQueryItem("language", lang);

        extra->setProperty("page", page);
        extra->setProperty("surah", querySpecific.value("surah"));
        extra->setProperty("ayah", querySpecific.value("ayah"));

        url = QUrl(QString("https://api.quran.com/api/v4/verses/by_chapter/%1").arg(querySpecific.value("surah").toString()));
    }
    else if (query == QURAN_AUDIO) {
        QDir      audioBaseDir(QStandardPaths::writableLocation(QStandardPaths::CacheLocation) + QDir::separator() + "audio");
        QFileInfo audioFile(audioBaseDir.absolutePath() + QDir::separator() + querySpecific.value("audio").toString());

        if (audioFile.exists()) {
            emit got(query, true, audioFile.canonicalFilePath());
            return;
        }

        FilesLastAccess filesLastAccess;
        if ((totalCachedBytes == 0) || (totalCachedBytes >= AUDIO_CACHE_MAX_BYTES)) {
            totalCachedBytes = 0;
            listFilesLastAccessRecursive(audioBaseDir, &totalCachedBytes, &filesLastAccess);
        }
        if (totalCachedBytes >= AUDIO_CACHE_MAX_BYTES) {
            std::sort(filesLastAccess.begin(), filesLastAccess.end(), [](FileLastAccess a, FileLastAccess b) {
                return a.lastRead < b.lastRead;
            });

            int cnt = 0;
            while (!filesLastAccess.isEmpty() && (totalCachedBytes >= AUDIO_CACHE_MAX_BYTES / 2)) {
                QFile file(filesLastAccess.first().file);
                if (file.remove()) {
                    totalCachedBytes -= filesLastAccess.first().size;
                }
                filesLastAccess.removeFirst();
                cnt++;
            }

            writeToErrorLog(QString("%1 files removed from audio cache reduced").arg(cnt));
        }

        extra->setProperty("file", audioFile.absoluteFilePath());
        url = QUrl(QString("https://verses.quran.com/%1").arg(querySpecific.value("audio").toString()));
    }

    url.setQuery(queryString);

    QNetworkRequest request(url);
    request.setRawHeader("User-Agent", QCoreApplication::instance()->applicationName().toUtf8());
    request.setOriginatingObject(extra);

    if ((query == QURAN_SUWAR) || (query == QURAN_AYAT)) {
        request.setRawHeader("Accept", "application/json");
    }

    writeToErrorLog(QString("Querying %1").arg(url.host()));
    networkAccessManager->get(request);
}


void NetworkingAndFiles::listFilesLastAccessRecursive(QDir dir, qint64* totalSize, FilesLastAccess* filesLastAccess)
{
    if (!dir.exists()) {
        return;
    }

    for(uint i = 0; i < dir.count(); i++) {
        if (dir[i].startsWith(".")) {
            continue;
        }

        QFileInfo entry(dir.absoluteFilePath(dir[i]));

        if (entry.isDir()) {
            listFilesLastAccessRecursive(QDir(entry.absoluteFilePath()), totalSize, filesLastAccess);
            continue;
        }

        *totalSize += entry.size();
        filesLastAccess->append({ entry.absoluteFilePath(), entry.lastRead().toSecsSinceEpoch(), entry.size() });
    }
}


void NetworkingAndFiles::networkFinished(QNetworkReply *reply)
{
    QVariant result;
    QUERIES  query = QUERY_UNKNOWN;

    if (reply->url() == reply->request().url()) {
        writeToErrorLog(QString("Query of %1 finished").arg(reply->request().url().host()));
    }
    else {
        writeToErrorLog(QString("Query of %1 (redirected to %2) finished").arg(reply->request().url().host(), reply->url().host()));
    }

    if (reply->error() != QNetworkReply::NoError) {
        QString msg = QString("Network error code %1").arg(reply->error());
        writeToErrorLog(msg);
        if (reply->errorString().length()) {
            writeToErrorLog(reply->errorString());
        }
        emit uiMsg(msg, false);
        emit got(query, false, result);

        if (isKiosk && (!repeatAfterConnected.url().isValid())) {
            QNetworkRequest request(reply->request());

            QObject *new_extra = new QObject();
            QObject *old_extra = reply->request().originatingObject();
            if (old_extra != nullptr) {
                QList<QByteArray> properties = old_extra->dynamicPropertyNames();
                foreach(QByteArray property, properties) {
                    new_extra->setProperty(property, old_extra->property(property));
                }
            }
            request.setOriginatingObject(new_extra);

            QList<QByteArray> rawHeaders = reply->request().rawHeaderList();
            foreach(QByteArray rawHeader, rawHeaders) {
                request.setRawHeader(rawHeader, reply->request().rawHeader(rawHeader));
            }

            checkConnection();
        }

        reply->deleteLater();
        return;
    }
    if (reply->attribute(QNetworkRequest::HttpStatusCodeAttribute) != 200) {
        QString msg = QString("HTTP status code %1 received").arg(reply->attribute(QNetworkRequest::HttpStatusCodeAttribute).toString());
        writeToErrorLog(msg);
        emit uiMsg(msg, false);
        emit got(query, false, result);
        reply->deleteLater();
        return;
    }

    QVariantMap extraData;

    QObject *extra = reply->request().originatingObject();
    if (extra != nullptr) {
        QVariant value = extra->property("queryId");
        if (value.isValid()) {
            query = static_cast<QUERIES>(value.toInt());
        }
        if (query == LOCATION_TIMEZONE) {
            foreach(QString property, QStringList({ "latitude", "longitude", "country", "countryCode", "region", "city" })) {
                extraData.insert(property, extra->property(property.toUtf8()));
            }
        }
        else if (query == QURAN_SUWAR) {
            extraData.insert("language", extra->property("language"));
        }
        else if (query == QURAN_AYAT) {
            foreach(QString property, QStringList({ "language", "page", "surah", "ayah" })) {
                extraData.insert(property, extra->property(property.toUtf8()));
            }
        }
        else if (query == QURAN_AUDIO) {
            extraData.insert("file", extra->property("file"));
        }
        delete extra;
    }
    if (query == QUERY_UNKNOWN) {
        writeToErrorLog("Unknown query");
        emit uiMsg("Unknown query", false);
        emit got(query, false, result);
        return;
    }

    QSettings settings;

    if (query == QUERY_LOCATION) {
        QJsonDocument jsonDoc = QJsonDocument::fromJson(reply->readAll());
        reply->deleteLater();

        if (!jsonDoc.isObject()) {
            writeToErrorLog("JSON parser error");
            emit uiMsg("JSON parser error", false);
            emit got(query, false, result);
            return;
        }
        if (jsonDoc.object().contains("success") && !jsonDoc.object().value("success").toBool()) {
            if (jsonDoc.object().contains("error") && jsonDoc.object().value("error").toObject().contains("code")) {
                QString msg = QString("IPstack error: %1").arg(jsonDoc.object().value("error").toObject().value("code").toString());
                writeToErrorLog(msg);
                emit uiMsg(msg, false);
                emit got(query, false, result);
            }
            else {
                writeToErrorLog("IPstack error");
                emit uiMsg("IPstack error", false);
                emit got(query, false, result);
            }
            if (jsonDoc.object().value("error").toObject().contains("info")) {
                writeToErrorLog(jsonDoc.object().value("error").toObject().value("info").toString());
            }
            return;
        }

        QString country      = jsonDoc.object().value("country_name").toString();
        QString region       = jsonDoc.object().value("region_name").toString();
        QString city         = jsonDoc.object().value("city").toString();
        QString timeZone     = jsonDoc.object().value("time_zone").toObject().value("id").toString();
        QString timeZoneCode = jsonDoc.object().value("time_zone").toObject().value("code").toString();

        settings.setValue("locationAuto/latitude",     jsonDoc.object().value("latitude").toDouble());
        settings.setValue("locationAuto/longitude",    jsonDoc.object().value("longitude").toDouble());
        settings.setValue("locationAuto/country",      country);
        settings.setValue("locationAuto/countryCode",  jsonDoc.object().value("country_code").toString());
        settings.setValue("locationAuto/region",       region);
        settings.setValue("locationAuto/city",         city);
        settings.setValue("locationAuto/timeZone",     timeZone);
        settings.setValue("locationAuto/timeZoneCode", timeZoneCode);
        settings.sync();

        emit got(query, true, jsonDoc.toVariant());
        return;
    }

    if ((query == QUERY_COUNTRIES) || (query == QUERY_REGIONS)) {
        QList<GeoPlace>    places;
        GeoPlace           place;
        QRegularExpression adminCode1Regexp("(\\w+)}?$");

        QString currentElement = "";
        int     level          = 0;

        QXmlStreamReader xmlStreamReader(reply);
        while (!xmlStreamReader.atEnd()) {
            QXmlStreamReader::TokenType tokenType = xmlStreamReader.readNext();
            if (tokenType == QXmlStreamReader::StartElement) {
                currentElement = xmlStreamReader.name().toString();
                level++;

                if (((query == QUERY_COUNTRIES) && (currentElement.compare("country") == 0)) || ((query == QUERY_REGIONS) && (currentElement.compare("geoname") == 0))) {
                    place.geonameId = "";
                    place.name      = "";
                    place.code      = "";
                }
            }
            else if (tokenType == QXmlStreamReader::Characters) {
                QString text = xmlStreamReader.text().toString();
                if (currentElement.compare("geonameId") == 0) {
                    place.geonameId = text;
                }
                else if ((query == QUERY_COUNTRIES) && (currentElement.compare("countryName") == 0)) {
                    place.name = text;
                }
                else if ((query == QUERY_REGIONS) && (currentElement.compare("name") == 0)) {
                    place.name = text;
                }
                else if ((query == QUERY_COUNTRIES) && (currentElement.compare("countryCode") == 0)) {
                    place.code = text;
                }
                else if ((query == QUERY_REGIONS) && (currentElement.compare("adminCode1") == 0)) {
                    place.code = text;
                }
            }
            else if (tokenType == QXmlStreamReader::EndElement) {
                if ((level == 2) && (!place.geonameId.isEmpty())) {
                    if (query == QUERY_REGIONS) {
                        QRegularExpressionMatch codeMatch = adminCode1Regexp.match(place.code);
                        if (codeMatch.hasMatch()) {
                            place.code = codeMatch.captured(1);
                            places.append(place);
                        }
                    }
                    else {
                        places.append(place);
                    }
                }

                currentElement = "";
                level--;
            }
        }

        reply->deleteLater();

        if (xmlStreamReader.hasError()) {
            QString msg = QString("XML parser error A%1").arg(xmlStreamReader.error());
            writeToErrorLog(msg);
            if (xmlStreamReader.errorString().length()) {
                writeToErrorLog(xmlStreamReader.errorString());
            }
            writeToErrorLog(reply->readAll());
            emit uiMsg(msg, false);
            emit got(query, false, result);
            return;
        }

        std::sort(places.begin(), places.end(), [](GeoPlace a, GeoPlace b) {
            return a.name.compare(b.name, Qt::CaseInsensitive) < 0;
        });

        QVariantList placesList;
        foreach(GeoPlace p, places) {
            placesList.append(QVariantMap({{ "geoname_id", p.geonameId }, { "name", p.name }, { "code", p.code }}));
        }

        if (query == QUERY_COUNTRIES) {
            QVariantMap countriesMap;

            countriesMap.insert("sel_code", settings.value(QString("%1/countryCode").arg(settings.childGroups().contains("locationCustom") ? "locationCustom" : "locationAuto"), ""));
            countriesMap.insert("countries", placesList);

            emit got(query, true, countriesMap);
        }
        else if (query == QUERY_REGIONS) {
            placesList.prepend(QVariantMap({{ "geoname_id", "-" }, { "name", tr("Search Entire Country") }, { "code", "-" }}));
            emit got(query, true, placesList);
        }
        return;
    }

    if ((query == LOCATION_SEARCH_POSTALCODE) || (query == LOCATION_SEARCH_FREETEXT)) {
        QVariantList locations;
        QVariantMap  location;
        QString      currentElement = "";
        int          level          = 0;

        QXmlStreamReader xmlStreamReader(reply);
        while (!xmlStreamReader.atEnd()) {
            QXmlStreamReader::TokenType tokenType = xmlStreamReader.readNext();
            if (tokenType == QXmlStreamReader::StartElement) {
                currentElement = xmlStreamReader.name().toString();
                level++;

                if (((query == LOCATION_SEARCH_FREETEXT) && (currentElement.compare("geoname") == 0)) || ((query == LOCATION_SEARCH_POSTALCODE) && (currentElement.compare("code") == 0))) {
                    location.clear();
                }
            }
            else if (tokenType == QXmlStreamReader::Characters) {
                QString text = xmlStreamReader.text().toString();
                if (currentElement.compare("name") == 0) {
                    location.insert("city", text);
                }
                else if (currentElement.compare("countryCode") == 0) {
                    location.insert("country_code", text);
                }
                else if (currentElement.compare("countryName") == 0) {
                    location.insert("country", text);
                }
                else if (currentElement.compare("adminName1") == 0) {
                    location.insert("region", text);
                }
                else if (currentElement.compare("lat") == 0) {
                    location.insert("latitude", text);
                }
                else if (currentElement.compare("lng") == 0) {
                    location.insert("longitude", text);
                }
                else if (currentElement.compare("timezone") == 0) {
                    location.insert("time_zone", text);
                }
            }
            else if (tokenType == QXmlStreamReader::EndElement) {
                if ((level == 2) && (!location.isEmpty())) {
                    QStringList displayParts;

                    if (location.contains("city")) {
                        displayParts.append(location.value("city").toString());
                    }
                    if (location.contains("region")) {
                        displayParts.append(location.value("region").toString());
                    }
                    if (location.contains("country")) {
                        displayParts.append(location.value("country").toString());
                    }
                    if ((displayParts.isEmpty() || !location.contains("country")) && location.contains("country_code")) {
                        displayParts.append(location.value("country_code").toString());
                    }

                    if (!displayParts.isEmpty()) {
                        location.insert("display_str", displayParts.join(", "));
                        locations.append(location);
                    }
                }

                currentElement = "";
                level--;
            }
        }

        reply->deleteLater();

        if (xmlStreamReader.hasError()) {
            QString msg = QString("XML parser error B%1").arg(xmlStreamReader.error());
            writeToErrorLog(msg);
            if (xmlStreamReader.errorString().length()) {
                writeToErrorLog(xmlStreamReader.errorString());
            }
            emit uiMsg(msg, false);
            emit got(query, false, result);
            return;
        }

        emit got(query, true, locations);
        return;
    }

    if (query == LOCATION_TIMEZONE) {
        QString timeZone;
        QString currentElement = "";

        QXmlStreamReader xmlStreamReader(reply);
        while (!xmlStreamReader.atEnd()) {
            QXmlStreamReader::TokenType tokenType = xmlStreamReader.readNext();
            if (tokenType == QXmlStreamReader::StartElement) {
                currentElement = xmlStreamReader.name().toString();
            }
            else if (tokenType == QXmlStreamReader::Characters) {
                if (currentElement.compare("timezoneId") == 0) {
                    timeZone = xmlStreamReader.text().toString();
                }
            }
            else if (tokenType == QXmlStreamReader::EndElement) {
                currentElement = "";
            }
        }

        reply->deleteLater();

        if (xmlStreamReader.hasError()) {
            QString msg = QString("XML parser error C%1").arg(xmlStreamReader.error());
            writeToErrorLog(msg);
            if (xmlStreamReader.errorString().length()) {
                writeToErrorLog(xmlStreamReader.errorString());
            }
            emit uiMsg(msg, false);
            emit got(query, false, result);
            return;
        }
        if (timeZone.isEmpty()) {
            writeToErrorLog("Time zone not found");
            emit uiMsg("Time zone not found", false);
            emit got(query, false, result);
            return;
        }
        if (extraData.isEmpty()) {
            writeToErrorLog("Internal error");
            emit uiMsg("Internal error", false);
            emit got(query, false, result);
            return;
        }

        QString timeZoneCode;
        QTimeZone tz(timeZone.toUtf8());
        if (tz.isValid()) {
            QDateTime now = QDateTime::currentDateTimeUtc();
            now.toTimeZone(tz);
            timeZoneCode = tz.displayName(now, QTimeZone::ShortName);
        }
        if (timeZoneCode.isEmpty()) {
            QDateTime now = QDateTime::currentDateTime();
            timeZone      = now.timeZone().displayName(now, QTimeZone::LongName);
            timeZoneCode  = now.timeZone().displayName(now, QTimeZone::ShortName);
        }

        settings.setValue("locationCustom/latitude",     extraData.value("latitude").toDouble());
        settings.setValue("locationCustom/longitude",    extraData.value("longitude").toDouble());
        settings.setValue("locationCustom/country",      extraData.value("country").toString());
        settings.setValue("locationCustom/countryCode",  extraData.value("countryCode").toString());
        settings.setValue("locationCustom/region",       extraData.value("region").toString());
        settings.setValue("locationCustom/city",         extraData.value("city").toString());
        settings.setValue("locationCustom/timeZone",     timeZone);
        settings.setValue("locationCustom/timeZoneCode", timeZoneCode);

        settings.sync();

        extraData.insert("timeZone",     timeZone);
        extraData.insert("timeZoneCode", timeZoneCode);

        emit got(query, true, extraData);
        return;
    }

    if (query == QURAN_SUWAR) {
        QJsonDocument jsonDoc = QJsonDocument::fromJson(reply->readAll());
        reply->deleteLater();

        if (!jsonDoc.isObject()) {
            writeToErrorLog("JSON parser error");
            emit uiMsg("JSON parser error", true);
            emit got(query, false, result);
            return;
        }
        if (!jsonDoc.object().contains("chapters")) {
            writeToErrorLog("quran.com error");
            emit uiMsg("quran.com error", true);
            emit got(query, false, result);
            return;
        }
        QJsonArray chapters = jsonDoc.object().value("chapters").toArray();

        if (extraData.isEmpty()) {
            writeToErrorLog("Internal error");
            emit uiMsg("Internal error", true);
            emit got(query, false, result);
            return;
        }
        QLocale languageLocale(extraData.value("language").toString());

        QVariantList suwar;

        bool chapterError = false;
        foreach(QJsonValue chapter, chapters) {
            QJsonObject chapterObj = chapter.toObject();
            if (chapterObj.isEmpty()) {
                if (!chapterError) {
                    chapterError = true;
                    writeToErrorLog("Unable to parse surah info 101");
                    emit uiMsg("Unable to parse surah info 101", true);
                    emit got(query, false, result);
                }
                continue;
            }

            int id        = chapterObj.value("id").toInt();
            int ayatCount = chapterObj.value("verses_count").toInt();

            if ((id == 0) || (ayatCount == 0)) {
                if (!chapterError) {
                    chapterError = true;
                    writeToErrorLog("Unable to parse surah info 102");
                    emit uiMsg("Unable to parse surah info 102", true);
                    emit got(query, false, result);
                }
                continue;
            }

            QString name;
            if (languageLocale.script() == QLocale::ArabicScript) {
                name = chapterObj.value("name_arabic").toString();
            }
            else {
                name = chapterObj.value("name_simple").toString();
            }

            QString number = QString("%1").arg(id);
            if (languageLocale.language() == QLocale::Persian) {
                number = Globals::convertNumbersToFarsi(number);
                if (name.isEmpty()) {
                    name = number;
                }
                else {
                    name.append(" ");
                    name.append(number);
                }
            }
            else if (languageLocale.script() == QLocale::ArabicScript) {
                number = Globals::convertNumbersToEastern(number);
                if (name.isEmpty()) {
                    name = number;
                }
                else {
                    name.append(" ");
                    name.append(number);
                }
            }
            else {
                if (name.isEmpty()) {
                    name = number;
                }
                else {
                    name.prepend(". ");
                    name.prepend(number);
                }
            }

            suwar.append(QVariantMap({{ "id", id }, { "ayat", ayatCount}, { "name", name }}));
        }

        if (!suwar.isEmpty()) {
            quranCacheSave(quranCacheFileName(QURAN_SUWAR, extraData.value("language").toString()), suwar);
            emit got(query, true, suwar);
        }
        return;
    }

    if (query == QURAN_AYAT) {
        QJsonDocument jsonDoc = QJsonDocument::fromJson(reply->readAll());
        reply->deleteLater();

        if (!jsonDoc.isObject()) {
            writeToErrorLog("JSON parser error");
            emit uiMsg("JSON parser error", true);
            emit got(query, false, result);
            return;
        }
        if (!jsonDoc.object().contains("verses")) {
            writeToErrorLog("quran.com error");
            emit uiMsg("quran.com error", true);
            emit got(query, false, result);
            return;
        }
        QJsonArray verses = jsonDoc.object().value("verses").toArray();

        if (extraData.isEmpty()) {
            writeToErrorLog("Internal error");
            emit uiMsg("Internal error", true);
            emit got(query, false, result);
            return;
        }

        int          selectedAyahNumber = extraData.value("ayah").toInt();
        QVariantMap  selectedAyah;
        QVariantList ayat;

        bool verseError = false;
        foreach(QJsonValue verse, verses) {
            QJsonObject verseObj = verse.toObject();
            if (verseObj.isEmpty()) {
                if (!verseError) {
                    verseError = true;
                    writeToErrorLog("Unable to parse surah info 201");
                    emit uiMsg("Unable to parse surah info 201", true);
                    emit got(query, false, result);
                }
                continue;
            }

            QVariantMap ayah = {
                { "id",                  verseObj.value("id") },
                { "verse_number",        verseObj.value("verse_number") },
                { "text_uthmani_simple", verseObj.value("text_uthmani_simple") },
                { "audio",               verseObj.value("audio").toObject().value("url").toString() }
            };

            QVariantList words;
            if (verseObj.contains("words")) {
                QJsonArray wordsArray = verseObj.value("words").toArray();
                if (wordsArray.isEmpty()) {
                    if (!verseError) {
                        verseError = true;
                        writeToErrorLog("Unable to parse surah info 202");
                        emit uiMsg("Unable to parse surah info 202", true);
                        emit got(query, false, result);
                    }
                    continue;
                }

                foreach(QJsonValue word, wordsArray) {
                    QJsonObject wordObj = word.toObject();
                    if (wordObj.isEmpty()) {
                        if (!verseError) {
                            verseError = true;
                            writeToErrorLog("Unable to parse word info");
                            emit uiMsg("Unable to parse word info", true);
                            emit got(query, false, result);
                        }
                        continue;
                    }

                    QVariantMap kalima = {
                        { "id",          wordObj.value("id")          },
                        { "position",    wordObj.value("position")    },
                        { "page_number", wordObj.value("page_number") },
                        { "code_v2",     wordObj.value("code_v2")     },
                    };
                    words.append(kalima);
                }
            }

            ayah.insert("kalimaat", words);
            ayat.append(ayah);

            if (verseObj.value("verse_number").toInt() == selectedAyahNumber) {
                selectedAyah = ayah;
            }
        }

        if (!ayat.isEmpty()) {
            quranCacheSave(quranCacheFileName(QURAN_AYAT, extraData.value("language").toString(), extraData.value("page").toInt(), QVariantMap({{ "surah", extraData.value("surah") }})), ayat);
            emit got(query, true, QVariantMap({ {"surah", extraData.value("surah")}, { "ayat", ayat }, { "selectedAyah", selectedAyah } }));
        }
        return;
    }

    if (query == QURAN_AUDIO) {
        QFile cacheFile(extraData.value("file").toString());

        QDir dir;
        if (!dir.mkpath(QFileInfo(cacheFile.fileName()).absolutePath())) {
            writeToErrorLog(QString("Error while saving '%1': Unable to create audio cache directory").arg(cacheFile.fileName()));
            emit uiMsg("Unable to create audio cache directory", true);
            emit got(query, false, result);
            return;
        }

        if (!cacheFile.open(QIODevice::WriteOnly)) {
            writeToErrorLog(QString("Error while saving '%1': Unable to create audio cache file").arg(cacheFile.fileName()));
            emit uiMsg("Unable to create audio cache file", true);
            emit got(query, false, result);
            return;
        }

        QByteArray buffer = reply->read(DOWNLOAD_BUFFER_BYTES);
        while (!reply->isFinished() || (buffer.size() > 0)) {
            cacheFile.write(buffer);
            buffer = reply->read(DOWNLOAD_BUFFER_BYTES);
        }

        cacheFile.close();
        reply->deleteLater();

        emit got(query, true, cacheFile.fileName());
        return;
    }
}


QString NetworkingAndFiles::quranCacheFileName(QUERIES query, QString language, int page, QVariantMap querySpecific)
{
    // ATTENTION! Don't forget to increase cache version number if any of the query parameters changes! This includes QURAN_QUERY_PAGE_SIZE.
    QString baseName = QString("%1_v%2_%3_%4").arg(query).arg(QURAN_QUERY_CACHE_VER).arg(language).arg(page);
    if (querySpecific.size()) {
        QStringList items;
        foreach(QString key, querySpecific.keys()) {
            items.append(QString("%1=%2").arg(key).arg(querySpecific.value(key).toString()));
        }
        baseName.append("_");
        baseName.append(items.join("&"));
    }
    baseName.append(".qqc");

    if (baseName.length() > 255) {   // max filename length in ext4 filesystem
        return "";
    }

    return QString("%1%2%3").arg(QStandardPaths::writableLocation(QStandardPaths::CacheLocation), QDir::separator(), baseName);
}


QVariantList NetworkingAndFiles::quranCacheLoad(QString cacheFileName)
{
    QFile cacheFile(cacheFileName);
    QVariantList retval;

    if (!cacheFile.exists()) {
        return retval;
    }

    if (!cacheFile.open(QIODevice::ReadOnly)) {
        writeToErrorLog(QString("Error while loading '%1': Unable to open cache file").arg(cacheFile.fileName()));
        emit uiMsg("Unable to open cache file", true);
        emit got(QUERY_UNKNOWN, false, 0);
        return retval;
    }

    QJsonParseError error;
    QJsonDocument json = QJsonDocument::fromJson(cacheFile.readAll(), &error);
    cacheFile.close();
    if (error.error != QJsonParseError::NoError) {
        writeToErrorLog(QString("Error while loading '%1': Unable to parse cached data 101").arg(cacheFile.fileName()));
        if (error.errorString().length()) {
            writeToErrorLog(error.errorString());
        }
        emit uiMsg("Unable to parse cached data 101", true);
        emit got(QUERY_UNKNOWN, false, 0);
        return retval;
    }
    if (!json.isArray()) {
        writeToErrorLog(QString("Error while loading '%1': Unable to parse cached data 102").arg(cacheFile.fileName()));
        emit uiMsg("Unable to parse cached data 102", true);
        emit got(QUERY_UNKNOWN, false, 0);
        return retval;
    }

    return json.array().toVariantList();
}


void NetworkingAndFiles::quranCacheSave(QString cacheFileName, QVariantList data)
{
    QFile cacheFile(cacheFileName);
    QJsonDocument json(QJsonArray::fromVariantList(data));

    QDir dir;
    if (!dir.mkpath(QFileInfo(cacheFileName).absolutePath())) {
        writeToErrorLog(QString("Error while saving '%1': Unable to create cache directory").arg(cacheFile.fileName()));
        emit uiMsg("Unable to create cache directory", true);
        emit got(QUERY_UNKNOWN, false, 0);
        return;
    }

    if (!cacheFile.open(QIODevice::WriteOnly)) {
        writeToErrorLog(QString("Error while saving '%1': Unable to create cache file").arg(cacheFile.fileName()));
        emit uiMsg("Unable to create cache file", true);
        emit got(QUERY_UNKNOWN, false, 0);
        return;
    }

    cacheFile.write(json.toJson(QJsonDocument::Compact));
    cacheFile.close();
}


void NetworkingAndFiles::run()
{
    networkAccessManager = new QNetworkAccessManager();
    mutex                = new QMutex();
    utilsProcess         = new QProcess();

    networkAccessManager->setRedirectPolicy(QNetworkRequest::NoLessSafeRedirectPolicy);
    connect(networkAccessManager, &QNetworkAccessManager::finished, this, &NetworkingAndFiles::networkFinished);

    connect(utilsProcess, QOverload<int, QProcess::ExitStatus>::of(&QProcess::finished), this, &NetworkingAndFiles::utilsFinished);
    connect(utilsProcess, &QProcess::errorOccurred,                                      this, &NetworkingAndFiles::utilsError);
    connect(utilsProcess, &QProcess::stateChanged,                                       this, &NetworkingAndFiles::utilsStateChanged);
}


void NetworkingAndFiles::setUILnaguage(QString uiLanguage)
{
    mutex->lock();
    this->uiLanguage = uiLanguage;
    mutex->unlock();
}


void NetworkingAndFiles::utilsFinished(int exitCode, QProcess::ExitStatus exitStatus)
{
    if (exitStatus != QProcess::NormalExit) {
        return;   // error message already sent in utilsError
    }

    if ((exitCode == 1) || (exitCode == 2)) {
        QString errMsg = utilsProcess->readAllStandardError().trimmed();
        if (errMsg.length()) {
            writeToErrorLog(errMsg);
            emit uiMsg(errMsg, false);
        }
        emit showWifiBusy(false);
        if (!repeatAfterConnected.url().isEmpty()) {
            repeatAfterConnected = QNetworkRequest();
        }
        return;
    }

    QString basename = QFileInfo(utilsProcess->program()).baseName();

    if (basename.compare("tef_inet_is_accessible") == 0) {
        if ((exitCode == 255) && (!wifiRetryTimestamp || (QDateTime::currentSecsSinceEpoch() - wifiRetryTimestamp > WIFI_RETRY_DELAY))) {
            wifiRetryTimestamp = QDateTime::currentSecsSinceEpoch();
            emit uiMsg("", false);
            execUtil("tef_wifi_is_connected");
        }
        return;
    }

    if (basename.compare("tef_wifi_is_connected") == 0) {
        if (exitCode == 255) {
            emit showWifiBusy(true);
            emit showWifiConnect(QStringList());
            execUtil("tef_wifi_on");
        }
        else {
            emit showWifiBusy(true);
            emit showWifiConnect(QStringList());
            execUtil("tef_wifi_list");
        }
        return;
    }

    if (basename.compare("tef_wifi_on") == 0) {
        execUtil("tef_wifi_list");
        return;
    }

    if (basename.compare("tef_wifi_list") == 0) {
        emit showWifiBusy(false);
        emit showWifiConnect(QString(utilsProcess->readAllStandardOutput().trimmed()).split('\n'));
        wifiRetryTimestamp = QDateTime::currentSecsSinceEpoch();
        return;
    }

    if (basename.compare("tef_wifi_connect") == 0) {
        if (repeatAfterConnected.url().isValid()) {
            networkAccessManager->get(repeatAfterConnected);
        }
        return;
    }
}


void NetworkingAndFiles::utilsError(QProcess::ProcessError error)
{
    QString errorStr = "Unknown error";

    switch (error) {
        case QProcess::FailedToStart:
            errorStr = "Failed to start";
            break;
        case QProcess::Crashed:
            errorStr = "Crashed";
            break;
        case QProcess::Timedout:
            errorStr = "Timed out";
            break;
        case QProcess::WriteError:
            errorStr = "Write error";
            break;
        case QProcess::ReadError:
            errorStr = "Read error";
    }

    QString basename = QFileInfo(utilsProcess->program()).baseName();
    writeToErrorLog(QString("%1 - %2").arg(basename, errorStr));
}


void NetworkingAndFiles::utilsStateChanged(QProcess::ProcessState state)
{
    QString stateStr = "Unknown";

    switch (state) {
        case QProcess::NotRunning:
            stateStr = "Not running";
            break;
        case QProcess::Starting:
            stateStr = "Starting";
            break;
        case QProcess::Running:
            stateStr = "Running";
    }

    QString basename = QFileInfo(utilsProcess->program()).baseName();
    writeToErrorLog(QString("%1 - %2").arg(basename, stateStr));
}


void NetworkingAndFiles::wifiConnect(QString ssid, QString psw)
{
    if (!ssid.length()) {
        if (repeatAfterConnected.url().isValid()) {
            repeatAfterConnected = QNetworkRequest();
        }
        return;
    }

    QStringList arguments(ssid);
    if (psw.length()) {
        arguments.append(psw);
    }

    execUtil("tef_wifi_connect", arguments);
}


void NetworkingAndFiles::writeToErrorLog(QString msg)
{
    if (!logging) {
        std::cerr << msg.append("\n").toUtf8().data();
        return;
    }

    if (errLogSize > ERR_LOG_MAX_SIZE) {
        errLog.close();

        if (!errLog.open(QIODevice::ReadOnly)) {
            logging = false;
            std::cerr << "Unable to open error log in read only mode\n";
            return;
        }

        QFile temp(QString("%1.temp").arg(errLog.fileName()));
        if (!temp.open(QIODevice::WriteOnly)) {
            errLog.close();
            logging = false;
            std::cerr << "Unable to create temporary error log\n";
            return;
        }

        char buf = 0;
        int  cnt = 0;
        errLog.seek(errLogSize / 2);
        while (buf != '\n') {
            if ((errLog.read(&buf, 1) <= 0) || (cnt >= 2048)) {
                break;
            }
            cnt++;
        }

        QByteArray largeBuf = errLog.read(1024 * 1024);
        while (largeBuf.size() > 0) {
            temp.write(largeBuf);
            largeBuf = errLog.read(1024 * 1024);
        }

        errLog.close();
        temp.close();

        if (!errLog.remove()) {
            logging = false;
            std::cerr << "Unable to delete error log\n";
            return;
        }
        if (!temp.rename(errLog.fileName())) {
            logging = false;
            std::cerr << "Unable to rename temporary error log\n";
            return;
        }

        errLogSize = errLog.size();
        if (!errLog.open(QIODevice::Append)) {
            logging = false;
            std::cerr << "Unable to reopen error log in read/write mode\n";
        }

        errLog.write(QString("%1 Error log was reduced to %2 bytes\n").arg(QDateTime::currentDateTime().toString("yyyy-MM-dd HH:mm:ss")).arg(errLogSize).toUtf8());
    }

    msg.prepend(QDateTime::currentDateTime().toString("yyyy-MM-dd HH:mm:ss") + " ");
    msg.append("\n");

    QByteArray utf8 = msg.toUtf8();
    errLog.write(utf8);
    errLog.flush();

    errLogSize += utf8.size();
}
