#ifndef NETWORKING_H
#define NETWORKING_H

#include <iostream>
#include <QByteArray>
#include <QCoreApplication>
#include <QDateTime>
#include <QFile>
#include <QFileInfo>
#include <QFileInfoList>
#include <QDir>
#include <QJsonArray>
#include <QJsonDocument>
#include <QJsonObject>
#include <QLocale>
#include <QList>
#include <QMutex>
#include <QNetworkAccessManager>
#include <QNetworkReply>
#include <QNetworkRequest>
#include <QObject>
#include <QProcess>
#include <QRegularExpression>
#include <QSettings>
#include <QStandardPaths>
#include <QTimeZone>
#include <QUrl>
#include <QUrlQuery>
#include <QVariant>
#include <QVariantList>
#include <QVariantMap>
#include <QXmlStreamReader>

#include "globals.h"
#include "secrets.h"


class NetworkingAndFiles : public QObject
{
    Q_OBJECT

    public:

        enum QUERIES {
            QUERY_UNKNOWN,
            QUERY_COUNTRIES,
            QUERY_LOCATION,
            QUERY_REGIONS,
            LOCATION_SEARCH_POSTALCODE,
            LOCATION_SEARCH_FREETEXT,
            LOCATION_TIMEZONE,
            QURAN_SUWAR,
            QURAN_AYAT,
            QURAN_AUDIO
        };

        static const int QURAN_QUERY_PAGE_SIZE = 10;

        NetworkingAndFiles(bool isKiosk);
        ~NetworkingAndFiles();

        void setUILnaguage(QString uiLanguage);

    private:

        struct GeoPlace {
            QString geonameId;
            QString name;
            QString code;
        };
        struct FileLastAccess {
            QString file;
            qint64  lastRead;
            qint64  size;
        };
        typedef QList<FileLastAccess> FilesLastAccess;

        static const qint64 AUDIO_CACHE_MAX_BYTES = static_cast<qint64>(1024) * 1024 * 1024 * 5;
        static const int    DOWNLOAD_BUFFER_BYTES = 1024 * 1024;
        static const qint64 ERR_LOG_MAX_SIZE      = 10 * 1024 * 1024;
        static const int    QURAN_QUERY_CACHE_VER = 1;
        static const int    SEARCH_LIMIT          = 99;
        static const qint64 WIFI_RETRY_DELAY      = 15 * 60;

        QString uiLanguage;
        qint64  totalCachedBytes;

        bool   isKiosk;
        bool   logging;
        QFile  errLog;
        qint64 errLogSize;
        qint64 wifiRetryTimestamp;

        QMutex                *mutex;
        QNetworkAccessManager *networkAccessManager;
        QNetworkRequest        repeatAfterConnected;

        QProcess *utilsProcess;

        QString      quranCacheFileName(QUERIES query, QString language, int page = 0, QVariantMap querySpecific = QVariantMap());
        QVariantList quranCacheLoad(QString cacheFileName);
        void         quranCacheSave(QString cacheFileName, QVariantList data);

        void listFilesLastAccessRecursive(QDir dir, qint64* totalSize, FilesLastAccess* filesLastAccess);

        void execUtil(QString basename);
        void execUtil(QString basename, QStringList arguments);


    public slots:

        void run();

        void checkConnection();
        void get(QUERIES query, int page = 0, QVariantMap querySpecific = QVariantMap());
        void wifiConnect(QString ssid, QString psw);
        void writeToErrorLog(QString msg);


    private slots:

        void networkFinished(QNetworkReply *reply);

        void utilsFinished(int exitCode, QProcess::ExitStatus exitStatus);
        void utilsError(QProcess::ProcessError error);
        void utilsStateChanged(QProcess::ProcessState state);


    signals:

        void got(QUERIES query, bool success, QVariant result);
        void uiMsg(QString msg, bool isTemporary);
        void showWifiConnect(QStringList ssids);
        void showWifiBusy(bool busy);

};

#endif // NETWORKING_H
