#ifndef PRAYERTIMES_H
#define PRAYERTIMES_H

#include <QDateTime>
#include <QHash>
#include <QList>
#include <QSettings>
#include <QTime>
#include <QTimeZone>
#include <QtMath>

#include "globals.h"


class PrayerTimes
{
    public:

        struct PrayerEvent {
            PRAYER_EVENTS prayerEvent;
            QTime         time;
            int           seconds;
            double        percent;
        };

        PrayerTimes();

        void    calculate();
        bool    isValid();
        QString lastError();

        QTime getFajr();
        QTime getShurooq();
        QTime getDhuhr();
        QTime getAsr();
        QTime getMaghrib();
        QTime getIsha();

        QTime getPrayer(PRAYER_EVENTS prayerEvent);

        PrayerEvent getPrevious();
        PrayerEvent getNext();


    private:

        struct HighLatitudeCorrection {
            double fajrCorrected;
            double ishaCorrected;
        };

        const double SUN_NATURAL_DEGREE         = 0.8333;
        const double EXTREME_LATITUDE_FAJR_ISHA = 18;
        const long   FASTING_MIN_HOURS          = 6;
        const long   FASTING_MAX_HOURS          = 18;
        const int    SHUROOQ_MINUTES            = 20;

        const double METHOD_EGYPT_FAJR        = 19.5;
        const double METHOD_EGYPT_ISHA        = 17.5;
        const double METHOD_NORTHAMERICA_FAJR = 15;
        const double METHOD_NORTHAMERICA_ISHA = 15;
        const double METHOD_PAKISTAN_FAJR     = 18;
        const double METHOD_PAKISTAN_ISHA     = 18;
        const double METHOD_SAUDI_FAJR        = 18.5;
        const double METHOD_SAUDI_ISHA        = 0;
        const double METHOD_IRAN_FAJR         = 16;
        const double METHOD_IRAN_ISHA         = 14;
        const double METHOD_EUANDWORLD_FAJR   = 18;
        const double METHOD_EUANDWORLD_ISHA   = 17;

        const double SCHOOL_HANBALI_MALIKI_SHAFI_SHADOW = 1;
        const double SCHOOL_HANAFI_SHADOW               = 2;

        bool    valid;
        QString errorMsg;

        QTime fajr;
        QTime shurooq;
        QTime dhuhr;
        QTime asr;
        QTime maghrib;
        QTime isha;

        QHash<PRAYER_EVENTS, QTime*> eventsToTime = {
            { FAJR,    &fajr    },
            { SHUROOQ, &shurooq },
            { DHUHR,   &dhuhr   },
            { ASR,     &asr     },
            { MAGHRIB, &maghrib },
            { ISHA,    &isha    }
        };

        CALCULATION_METHODS autoDetectMethod(QString countryCode);
        CALCULATION_SCHOOLS autoDetectSchool(QString countryCode);

        double getFajrAngle(int method);
        double getIshaAngle(int method);
        double getShadow(int school);

        double calculateDeclinationOfSun(int dayOfYear);
        double calculateEquationOfTime(int dayOfYear);
        double rad2deg(double what);
        double deg2rad(double what);
        double timeBetweenNoonAndSunAtAngle(double angle, double latitude, double declanation);
        double timeBetweenNoonAndShadowLenght(double shadowLength, double latitude, double declanation);
        double regularPrayerTime(double midday, double angle, double latitude, double declanationOfSun, int offsetMins);
        double afternoonPrayerTime(double midday, double shadowLength, double latitude, double declanationOfSun);

        HighLatitudeCorrection highLatitudeCorrection(int highLatitude, double fajrAngle, double ishaAngle, double shurooqFractional, double maghribFractional);
};

#endif // PRAYERTIMES_H
