#include "tefcacapplication.h"

TefCacApplication::TefCacApplication(int &argc, char **argv) : QGuiApplication(argc, argv)
{
    qRegisterMetaType<NetworkingAndFiles::QUERIES>("NetworkingAndFiles::QUERIES");
    qRegisterMetaType<NetworkingAndFiles::QUERIES>("QUERIES");

    adhanFajrIndex     = 0;
    adhanIndex         = 0;
    ayatMemCacheSurah  = 0;
    countriesQueried   = false;
    locationDataExists = false;
    networkingAndFiles = nullptr;
    playStartedByQuran = false;
    screenAfterSplash  = SCREEN_MAIN;
    soundPlayed        = false;
    suwarInitialized   = false;

    setOrganizationName("The Efendi Factor");
    setOrganizationDomain("efendif.com");
    setApplicationName("tef_cac");
    setApplicationDisplayName("Efendi Adhan Clock");
    setApplicationVersion(QString("%1.%2.%3").arg(VERSION_MAJOR).arg(VERSION_MINOR).arg(VERSION_SUB));
    setWindowIcon(QIcon("qrc:/images/waver.png"));

    networkingAndFilesThread.setObjectName("NetworkingAndFiles");

    connect(&backgroundTimer, &QTimer::timeout,                                         this, &TefCacApplication::swapBackground);
    connect(&mediaPlayer,     &QMediaPlayer::stateChanged,                              this, &TefCacApplication::mediaStateChanged);
    connect(&mediaPlayer,     QOverload<QMediaPlayer::Error>::of(&QMediaPlayer::error), this, &TefCacApplication::mediaError);

    connect(&hijriProcess, QOverload<int, QProcess::ExitStatus>::of(&QProcess::finished), this, &TefCacApplication::hijriFinished);
    connect(&hijriProcess, &QProcess::errorOccurred,                                      this, &TefCacApplication::hijriError);

    defLocale  = QLocale::system();
    translator = nullptr;
}


TefCacApplication::~TefCacApplication()
{
    mainTimer.stop();
    backgroundTimer.stop();

    networkingAndFilesThread.quit();
    networkingAndFilesThread.wait();

    if (translator != nullptr) {
        removeTranslator(translator);
        delete translator;
    }
}


void TefCacApplication::checkUpdatesButtonPressed()
{
    exit(EXIT_CODE_CHECK_UPDATES);
}


QVariantMap TefCacApplication::currentLocation()
{
    QSettings   settings;
    QVariantMap retVal;

    QString group = settings.childGroups().contains("locationCustom") ? "locationCustom" : "locationAuto";
    if (!settings.childGroups().contains(group)) {
        return retVal;
    }

    QStringList displayParts;

    QString temp = settings.value(QString("%1/city").arg(group), "").toString();
    if (!temp.isEmpty()) {
        retVal.insert("city", temp);
        displayParts.append(temp);
    }
    temp = settings.value(QString("%1/region").arg(group), "").toString();
    if (!temp.isEmpty()) {
        retVal.insert("region", temp);
        displayParts.append(temp);
    }
    temp = settings.value(QString("%1/country").arg(group), "").toString();
    if (!temp.isEmpty()) {
        retVal.insert("country", temp);
        displayParts.append(temp);
    }
    temp = settings.value(QString("%1/latitude").arg(group), "").toString();
    if (!temp.isEmpty()) {
        retVal.insert("latitude", temp);
    }
    temp = settings.value(QString("%1/longitude").arg(group), "").toString();
    if (!temp.isEmpty()) {
        retVal.insert("longitude", temp);
    }
    temp = settings.value(QString("%1/timeZone").arg(group), "").toString();
    if (!temp.isEmpty()) {
        retVal.insert("time_zone", temp);
    }
    temp = settings.value(QString("%1/timeZoneCode").arg(group), "").toString();
    if (!temp.isEmpty()) {
        retVal.insert("time_zone_code", temp);
    }
    temp = settings.value(QString("%1/countryCode").arg(group), "").toString();
    if (!temp.isEmpty()) {
        retVal.insert("country_code", temp);
        if (displayParts.isEmpty()) {
            displayParts.append(temp);
        }
    }

    if (!displayParts.isEmpty()) {
        retVal.insert("display_str", displayParts.join(", "));
    }

    return retVal;
}


bool TefCacApplication::envToBool(QString envVar)
{
    QString value = qgetenv(envVar.toUtf8().data());
    value = value.trimmed();

    if (value.length() < 1) {
        return false;
    }

    bool intOK    = false;
    int  intValue = value.toInt(&intOK);
    if (intOK && !intValue) {
        return false;
    }

    if (value.compare("false", Qt::CaseInsensitive) == 0) {
        return false;
    }

    return true;
}


QDateTime TefCacApplication::getCurrentDateTime()
{
    // TODO do a LOCATION_TIMEZONE query every now and then (like once a week or so)
    // just to keep up with changes in laws for example daylight saving usage

    if (!timeZoneCache.isValid()) {
        QSettings settings;

        QString group = settings.childGroups().contains("locationCustom") ? "locationCustom" : "locationAuto";

        QString timeZoneId = settings.value(QString("%1/timeZone").arg(group), "").toString();
        if (timeZoneId.isEmpty()) {
            return QDateTime::currentDateTime();
        }
        timeZoneCache = QTimeZone(timeZoneId.toUtf8());
    }

    QDateTime currentDateTime = QDateTime::currentDateTimeUtc().toTimeZone(timeZoneCache);

    return currentDateTime;
}


QString TefCacApplication::getUILanguage()
{
    QStringList uiLanguages = defLocale.uiLanguages();
    return uiLanguages.size() ? uiLanguages.at(0) : "en";
}


void TefCacApplication::handleErrorMsg(QString msg, bool isTemporary)
{
    if (msg.length()) {
        emit logErr(msg);
    }
    emit setErrorStr(msg, isTemporary);
}


void TefCacApplication::hijriError(QProcess::ProcessError error)
{
    Q_UNUSED(error);
    hijriFinished(999, QProcess::CrashExit);
}


void TefCacApplication::hijriFinished(int exitCode, QProcess::ExitStatus exitStatus)
{
    QSettings settings;
    QLocale::FormatType dateFormatType = settings.value("prefs/date_format_full", DEFAULT_DATE_FORMAT_FULL).toBool() ? QLocale::LongFormat : QLocale::ShortFormat;

    QDate nowDate = getCurrentDateTime().date();

    QString gregorian;
    QString hijri;
    QString shamsi;

    if (settings.value("prefs/gregorian_date_visible", DEFAULT_DATE_GREGORIAN_VISIBLE).toBool()) {
        QLocale locale = settings.value("prefs/gregorian_date_always_english", DEFAULT_DATE_GREGORIAN_ALWAYS_ENGLISH).toBool() ? QLocale(QLocale::English, QLocale::UnitedStates) : defLocale;
        gregorian = locale.toString(nowDate, locale.dateFormat(dateFormatType));
    }

    if ((exitCode == 0) && (exitStatus == QProcess::NormalExit)) {
        QString hijriOutput = hijriProcess.readAllStandardOutput();

        if (settings.value("prefs/hijri_date_visible", DEFAULT_DATE_HIJRI_VISIBLE).toBool()) {
            QRegularExpression      hijriRegexp(dateFormatType == QLocale::LongFormat ? "Hijri\\sfull:\\n(.*?)\\n" : "Hijri\\sshort:\\n(.*?)\\n");
            QRegularExpressionMatch hijriMatch = hijriRegexp.match(hijriOutput);
            if (hijriMatch.hasMatch()) {
                hijri = hijriMatch.captured(1);
            }
        }
        if (settings.value("prefs/shamsi_date_visible", DEFAULT_DATE_SHAMSI_VISIBLE).toBool()) {
            QRegularExpression      shamsiRegexp(dateFormatType == QLocale::LongFormat ? "Shamsi\\sfull:\\n(.*?)\\n" : "Shamsi\\sshort:\\n(.*?)\\n");
            QRegularExpressionMatch shamsiMatch = shamsiRegexp.match(hijriOutput);
            if (shamsiMatch.hasMatch()) {
                shamsi = shamsiMatch.captured(1);
            }
        }
    }

    if (hijri.isEmpty() && settings.value("prefs/hijri_date_visible", DEFAULT_DATE_HIJRI_VISIBLE).toBool()) {
        QLocale arabic = QLocale(QLocale::Arabic, QLocale::SaudiArabia);
        hijri = arabic.toString(nowDate, arabic.dateFormat(dateFormatType));
    }
    if (shamsi.isEmpty() && settings.value("prefs/shamsi_date_visible", DEFAULT_DATE_SHAMSI_VISIBLE).toBool()) {
        QLocale irani = QLocale(QLocale::Persian, QLocale::Iran);
        shamsi = irani.toString(nowDate, irani.dateFormat(dateFormatType));
    }

    emit setDate(gregorian, hijri, shamsi);
}


void TefCacApplication::installTranslation()
{
    if (translator != nullptr) {
        removeTranslator(translator);
        delete translator;
    }

    QSettings settings;
    QString   localeAbbr = settings.value("prefs/language", "-").toString();

    defLocale = localeAbbr.compare("-") == 0 ? QLocale::system() : QLocale(localeAbbr);

    emit setFont(
        settings.value("ui/font_family_latin", "sans"),
        settings.value("ui/font_family_arabic", "sans"),
        settings.value("ui/font_size", DEFAULT_FONT_SIZE),
        (settings.value("prefs/prayer_times_names", DEFAULT_PRAYER_NAMES).toInt() == PRAYER_NAMES_ARABIC) || (defLocale.script() == QLocale::ArabicScript),
        defLocale.script() == QLocale::ArabicScript
    );

    translator = new QTranslator();
    if (translator->load(defLocale, applicationName(), "_")) {
        installTranslator(translator);
    }
    else {
        delete translator;
        translator = nullptr;
    }
}


void TefCacApplication::mainTimerProc()
{
    QSettings settings;

    if (splash) {
        splash = false;

        if (settings.value("prefs/use_bkg", DEFAULT_USE_BKG).toBool()) {
            backgroundIndex = 0;
            swapBackground();
            backgroundTimer.start(BACKGROUND_DELAY_MS);
        }
        else {
            backgroundTimer.stop();
            emit hideBkg();
        }
        emit setScreen(screenAfterSplash);

        mainTimer.singleShot(50, this, &TefCacApplication::mainTimerProc);
        return;
    }

    updateScreen();
    mainTimer.singleShot(950, this, &TefCacApplication::mainTimerProc);
}


void TefCacApplication::mediaError(QMediaPlayer::Error error)
{
    playStartedByQuran = false;

    emit logErr(mediaPlayer.errorString());

    switch (error) {
        case QMediaPlayer::AccessDeniedError:
            handleErrorMsg("The media player reported an 'access denied' error", true);
            break;
        case QMediaPlayer::FormatError:
            handleErrorMsg("The media player reported a 'format' error", true);
            break;
        case QMediaPlayer::MediaIsPlaylist:
            handleErrorMsg("The media player reported a 'media is playlist' error", true);
            break;
        case QMediaPlayer::NetworkError:
            handleErrorMsg("The media player reported a 'network' error", true);
            break;
        case QMediaPlayer::NoError:
            handleErrorMsg("The media player reported a 'no error' message", true);
            break;
        case QMediaPlayer::ResourceError:
            handleErrorMsg("The media player reported a 'resource' error", true);
            break;
        case QMediaPlayer::ServiceMissingError:
            handleErrorMsg("The media player reported a 'service missing' error", true);
            break;
    }
}


void TefCacApplication::mediaStateChanged(QMediaPlayer::State state)
{
    emit setUIButtonRole(state == QMediaPlayer::PlayingState ? BUTTON_ROLE_STOP : BUTTON_ROLE_SETTINGS);

    if ((state == QMediaPlayer::StoppedState) && playStartedByQuran) {
        emit quranNextAyah(true);
    }
}


void TefCacApplication::networkGot(NetworkingAndFiles::QUERIES query, bool success, QVariant result)
{
    emit setLocationBusy(false);
    emit setQuranBusy(false);

    if (!success) {
        return;
    }

    QSettings settings;

    if (query == NetworkingAndFiles::QUERY_COUNTRIES) {
        emit setCountries(result.toMap());
        countriesQueried = true;
        return;
    }

    if (query == NetworkingAndFiles::QUERY_LOCATION) {
        QVariantMap results = result.toMap();

        timeZoneCache      = QTimeZone(results.value("time_zone").toMap().value("id").toString().toUtf8());
        locationDataExists = true;

        emit setLocation(results.value("city").toString(), results.value("region").toString(), results.value("country").toString(), results.value("time_zone").toMap().value("code").toString());
        updateScreen(true);

        return;
    }

    if (query == NetworkingAndFiles::QUERY_REGIONS) {
        emit setRegions(result.toList());
        return;
    }

    if ((query == NetworkingAndFiles::LOCATION_SEARCH_POSTALCODE) || (query == NetworkingAndFiles::LOCATION_SEARCH_FREETEXT)) {
        QVariantList locations = result.toList();
        QVariantMap  location  = currentLocation();

        if (!location.isEmpty()) {
            locations.prepend(location);
        }

        emit setSearchResults(locations);
        return;
    }

    if (query == NetworkingAndFiles::LOCATION_TIMEZONE) {
        QVariantMap results = result.toMap();

        timeZoneCache = QTimeZone(results.value("timeZone").toString().toUtf8());

        emit setLocation(results.value("city").toString(), results.value("region").toString(), results.value("country").toString(), results.value("timeZoneCode").toString());
        updateScreen(true);
        return;
    }

    if (query == NetworkingAndFiles::QURAN_SUWAR) {
        suwarInitialized = true;
        emit setQuranSuwar(result, settings.value("quran/surah", DEFAULT_SURAH).toInt(), settings.value("quran/ayah", DEFAULT_AYAH).toInt());
        return;
    }

    if (query == NetworkingAndFiles::QURAN_AYAT) {
        QVariantMap results = result.toMap();

        ayatMemCacheSurah = results.value("surah").toInt();
        ayatMemCache      = results.value("ayat").toList();
        if (results.contains("selectedAyah") && !results.value("selectedAyah").toMap().isEmpty()) {
            quranSetUIAyah(results.value("selectedAyah").toMap());
        }

        return;
    }

    if (query == NetworkingAndFiles::QURAN_AUDIO) {
        playStartedByQuran = true;
        playSound(QFileInfo(result.toString()));
        emit setQuranAudioToggleRole(true);
        return;
    }
}


void TefCacApplication::networkUIMsg(QString msg, bool isTemporary)
{
    emit setErrorStr(msg, isTemporary);
}


void TefCacApplication::networkWifiBusy(bool busy)
{
    emit setWifiBusy(busy);
}


void TefCacApplication::networkWifiLoginRequest(QStringList ssids)
{
    screenAfterSplash = SCREEN_WIFI;

    emit setScreen(SCREEN_WIFI);
    emit setWifiList(ssids);
}


void TefCacApplication::playAdhan(PRAYER_EVENTS prayer)
{
    QSettings settings;

    int sound = DEFAULT_PRAYER_SOUND;
    switch (prayer) {
        case FAJR:
            sound = settings.value("prefs/sound_fajr", DEFAULT_PRAYER_SOUND).toInt();
            break;
        case SHUROOQ:
            sound = settings.value("prefs/sound_shurooq", DEFAULT_SHUROOQ_SOUND).toInt();
            break;
        case DHUHR:
            sound = settings.value("prefs/sound_dhuhr", DEFAULT_PRAYER_SOUND).toInt();
            break;
        case ASR:
            sound = settings.value("prefs/sound_asr", DEFAULT_PRAYER_SOUND).toInt();
            break;
        case MAGHRIB:
            sound = settings.value("prefs/sound_maghrib", DEFAULT_PRAYER_SOUND).toInt();
            break;
        case ISHA:
            sound = settings.value("prefs/sound_isha", DEFAULT_PRAYER_SOUND).toInt();
    }

    playAdhanSound(prayer, sound);
}


void TefCacApplication::playAdhanSound(PRAYER_EVENTS prayer, int sound)
{
    QSettings settings;

    if (((prayer != SHUROOQ) && (sound == SILENT)) || ((prayer == SHUROOQ) && (sound == SHUROOQ_SILENT))) {
        return;
    }
    if (((prayer != SHUROOQ) && (sound == BEEP)) || ((prayer == SHUROOQ) && (sound == SHUROOQ_BEEP))) {
        playStartedByQuran = false;
        playSound(QFileInfo(QString("%1/beep.wav").arg(settings.value("media/sound_dir", DEFAULT_MEDIA_DIR).toString())));
        return;
    }

    QString dirName = (prayer == FAJR) ? "adhan_fajr" : "adhan";
    QString dirPath = settings.value(QString("media/%1_dir").arg(dirName), QString("%1/%2").arg(settings.value("media/sound_dir", DEFAULT_MEDIA_DIR).toString()).arg(dirName)).toString();

    QDir dir(dirPath);
    if (!dir.exists()) {
        emit logErr(QString("Adhan directory: %1").arg(dirPath));
        handleErrorMsg(QString("Directory for %1 not found").arg(dirName), false);
        return;
    }

    QFileInfoList sounds;
    QFileInfoList entries = dir.entryInfoList(QDir::Files | QDir::Readable);
    foreach (QFileInfo entry, entries) {
        if (entry.fileName().endsWith(".wav")) {
            sounds.append(entry);
        }
    }

    if (sounds.isEmpty()) {
        emit logErr(QString("Adhan directory: %1").arg(dirPath));
        handleErrorMsg("Adhan sound not found", false);
        return;
    }

    playStartedByQuran = false;

    std::sort(sounds.begin(), sounds.end(), [](QFileInfo a, QFileInfo b) {
        return a.baseName().compare(b.baseName(), Qt::CaseInsensitive) < 0;
    });

    if (prayer == FAJR) {
        if (adhanFajrIndex >= sounds.size()) {
            adhanFajrIndex = 0;
        }
        playSound(sounds.at(adhanFajrIndex));
        adhanFajrIndex++;
    }
    else {
        if (adhanIndex >= sounds.size()) {
            adhanIndex = 0;
        }
        playSound(sounds.at(adhanIndex));
        adhanIndex++;
    }
}


bool TefCacApplication::playSound(QFileInfo soundFile)
{
    if (soundFile.exists()) {
        mediaPlayer.stop();
        mediaPlayer.setMedia(QUrl::fromLocalFile(soundFile.absoluteFilePath()));
        mediaPlayer.play();
        return true;
    }

    emit logErr(QString("Sound file: %1").arg(soundFile.absoluteFilePath()));
    handleErrorMsg("Sound file doesn't exist", false);
    return false;
}


QString TefCacApplication::prayerName(PRAYER_EVENTS prayer, bool full)
{
    QSettings settings;

    PRAYER_NAMES prayerNames = static_cast<PRAYER_NAMES>(settings.value("prefs/prayer_times_names", DEFAULT_PRAYER_NAMES).toInt());
    bool         useSubh     = settings.value("prefs/prayer_times_subh", DEFAULT_PRAYER_NAMES_SUBH).toBool();

    if (prayerNames == PRAYER_NAMES_TRANSLITERATION) {
        switch (prayer) {
            case FAJR:
                if (full) {
                    return useSubh ? tr("Subh prayer") : tr("Fajr prayer");
                }
                return useSubh ? tr("Subh") : tr("Fajr");
            case SHUROOQ:
                return tr("Shurooq");
            case DHUHR:
                return full ? tr("Dhuhr prayer")   : tr("Dhuhr");
            case ASR:
                return full ? tr("Asr prayer")     : tr("Asr");
            case MAGHRIB:
                return full ? tr("Maghrib prayer") : tr("Maghrib");
            case ISHA:
                return full ? tr("Isha prayer")    : tr("Isha");
        }
    }

    if (prayerNames == PRAYER_NAMES_ARABIC) {
        QTranslator translator;
        QLocale     arabic(QLocale::Arabic);
        translator.load(arabic, applicationName(), "_");

        QString source;

        switch (prayer) {
            case FAJR:
                if (full) {
                    source = useSubh ? "Morning prayer" : "Dawn prayer";
                }
                else {
                    source = useSubh ? "Morning" : "Dawn";
                }
                break;
            case SHUROOQ:
                source = "Sunrise";
                break;
            case DHUHR:
                source = full ? "Noon prayer"      : "Noon";
                break;
            case ASR:
                source = full ? "Afternoon prayer" : "Afternoon";
                break;
            case MAGHRIB:
                source = full ? "Sunset prayer"    : "Sunset";
                break;
            case ISHA:
                source = full ? "Night prayer"     : "Night";
        }

        QString returnValue = translator.translate("TefCacApplication", source.toUtf8());
        if (returnValue.isEmpty()) {
            return source;
        }
        return returnValue;
    }

    if (prayerNames == PRAYER_NAMES_LOCAL) {
        switch (prayer) {
            case FAJR:
                if (full) {
                    return useSubh ? tr("Morning prayer") : tr("Dawn prayer");
                }
                return useSubh ? tr("Morning") : tr("Dawn");
            case SHUROOQ:
                return tr("Sunrise");
            case DHUHR:
                return full ? tr("Noon prayer")      : tr("Noon");
            case ASR:
                return full ? tr("Afternoon prayer") : tr("Afternoon");
            case MAGHRIB:
                return full ? tr("Sunset prayer")    : tr("Sunset");
            case ISHA:
                return full ? tr("Night prayer")     : tr("Night");
        }
    }

    return "?";
}


QString TefCacApplication::prayerTimeFormatted(QTime prayerTime, bool remainingtime)
{
    QSettings settings;

    QString timeStr;
    if (remainingtime) {
        int seconds = getCurrentDateTime().time().secsTo(prayerTime);
        timeStr = QString("%1:%2").arg(seconds / 60).arg(seconds % 60, 2, 10, QChar('0'));
    }
    else {
        bool time24 = settings.value("prefs/time_format_24", false).toBool();
        int  hours  = prayerTime.hour();
        if (!time24) {
            if (hours > 12) {
                hours -= 12;
            }
            if (hours == 0) {
                hours = 12;
            }
        }
        timeStr = QString("%1:%2").arg(hours, time24 ? 2 : 0, 10, QChar('0')).arg(prayerTime.minute(), 2, 10, QChar('0'));
    }

    if (settings.value("prefs/western_arabic_time_visible", DEFAULT_TIME_WESTERN_ARABIC_VISIBLE).toBool()) {
        return timeStr;
    }
    if (settings.value("prefs/eastern_arabic_time_visible", DEFAULT_TIME_EASTERN_ARABIC_VISIBLE).toBool()) {
        return Globals::convertNumbersToEastern(timeStr);
    }
    return Globals::convertNumbersToFarsi(timeStr);
}


void TefCacApplication::quranAudioTogglePressed(int role, int surah, int ayah)
{
    if (role == BUTTON_ROLE_STOP) {
        playStartedByQuran = false;
        mediaPlayer.stop();
        emit setQuranAudioToggleRole(false);
        return;
    }
    if (role != BUTTON_ROLE_PLAY) {
        return;
    }

    if (ayatMemCacheSurah != surah) {
        // TODO itt megáll az áják közötti létetés ha a következő szúrára lép
        return;
    }

    QVariantMap ayahObj = quranFindAyahInMemCache(ayah);
    if (ayahObj.isEmpty()) {
        return;
    }
    if (!ayahObj.contains("audio")) {
        return;
    }
    emit networkGet(NetworkingAndFiles::QURAN_AUDIO, 0, QVariantMap({{ "audio", ayahObj.value("audio") }}));
}


void TefCacApplication::quranAyahChanged(int surah, int ayah)
{
    QSettings settings;

    if (suwarInitialized) {
        settings.setValue("quran/surah", surah);
        settings.setValue("quran/ayah", ayah);
        settings.sync();
    }

    if (ayatMemCacheSurah == surah) {
        QVariantMap selectedAyah = quranFindAyahInMemCache(ayah);
        if (!selectedAyah.isEmpty()) {
            quranSetUIAyah(selectedAyah);
            return;
        }
    }

    emit networkGet(NetworkingAndFiles::QURAN_AYAT, (ayah - 1) / NetworkingAndFiles::QURAN_QUERY_PAGE_SIZE + 1, QVariantMap({ { "surah", surah }, { "ayah", ayah } }));
}


void TefCacApplication::quranButtonPressed()
{
    if (!suwarInitialized) {
        emit networkGet(NetworkingAndFiles::QURAN_SUWAR);
    }

    emit setScreen(SCREEN_QURAN);
}


QVariantMap TefCacApplication::quranFindAyahInMemCache(int ayahNumber)
{
    QVariantMap retval;

    bool errorShown = false;
    for(int i = 0; i < ayatMemCache.count(); i++) {
        QVariantMap ayahObj = ayatMemCache.value(i).toMap();
        if (ayahObj.isEmpty()) {
            if (!errorShown) {
                errorShown = true;
                handleErrorMsg("Unable to parse memory cached ayah info", true);
            }
            continue;
        }

        if (ayahObj.value("verse_number").toInt() == ayahNumber) {
            retval = ayahObj;
            break;
        }
    }

    return retval;
}


void TefCacApplication::quranSetUIAyah(QVariantMap ayah)
{
    if (fontsLoaded.count() > MAX_LOADED_FONTS) {
        QFontDatabase::removeAllApplicationFonts();
        fontsLoaded.clear();
    }

    QSettings settings;
    QString   ayahText       = ayah.value("text_uthmani_simple").toString();
    QString   ayahFontPage   = 0;
    QString   ayahFontFamily = "";

    QDir         fontDir(settings.value("media/v2_dir", QString("%1/ff_v2").arg(DEFAULT_MEDIA_DIR)).toString());
    QVariantList kalimaat = ayah.value("kalimaat").toList();
    if (fontDir.exists() && !kalimaat.isEmpty()) {
        QStringList words;
        for(int i = 0; i < kalimaat.count(); i++) {
            QVariantMap kalima = kalimaat.at(i).toMap();
            if (kalima.isEmpty() || (kalima.value("position").toInt() != i + 1)) {
                words.clear();
                break;
            }

            int page = kalima.value("page_number").toInt();
            if (!fontsLoaded.keys().contains(page)) {
                QFileInfo file(fontDir.filePath(QString("p%1.ttf").arg(page)));
                if (!file.exists()) {
                    words.clear();
                    break;
                }
                int id = QFontDatabase::addApplicationFont(file.canonicalFilePath());
                if (id < 0) {
                    words.clear();
                    break;
                }
                fontsLoaded.insert(page, id);
            }

            QStringList families = QFontDatabase::applicationFontFamilies(fontsLoaded.value(page));
            if (families.count() != 1) {
                words.clear();
                break;
            }

            if (ayahFontPage == 0) {
                ayahFontPage   = page;
                ayahFontFamily = families.first();
            }

            if (ayahFontPage != page) {
                words.append(QString("<div style='font-family: %1'>%2</div>").arg(families.first(), kalima.value("code_v2").toString()));
                continue;
            }
            words.append(kalima.value("code_v2").toString());
        }

        if (!words.isEmpty()) {
            ayahText = words.join(" ");
        }
    }

    emit setQuranAyah(ayahText, ayahFontFamily);
}


void TefCacApplication::setQmlApplicationEngine(QQmlApplicationEngine *qmlApplicationEngine)
{
    this->qmlApplicationEngine = qmlApplicationEngine;

    QQuickWindow *uiMainWindow = qobject_cast<QQuickWindow *>(this->qmlApplicationEngine->rootObjects().at(0));

    networkingAndFiles = new NetworkingAndFiles(envToBool("TEF_CAC_KIOSK_MODE"));
    networkingAndFiles->moveToThread(&networkingAndFilesThread);

    connect(&networkingAndFilesThread, &QThread::started,  networkingAndFiles, &NetworkingAndFiles::run);
    connect(&networkingAndFilesThread, &QThread::finished, networkingAndFiles, &NetworkingAndFiles::deleteLater);

    connect(this,               &TefCacApplication::networkGet,       networkingAndFiles, &NetworkingAndFiles::get);
    connect(this,               &TefCacApplication::logErr,           networkingAndFiles, &NetworkingAndFiles::writeToErrorLog);
    connect(this,               &TefCacApplication::wifiConnect,      networkingAndFiles, &NetworkingAndFiles::wifiConnect);
    connect(networkingAndFiles, &NetworkingAndFiles::got,             this,               &TefCacApplication::networkGot);
    connect(networkingAndFiles, &NetworkingAndFiles::uiMsg,           this,               &TefCacApplication::networkUIMsg);
    connect(networkingAndFiles, &NetworkingAndFiles::showWifiConnect, this,               &TefCacApplication::networkWifiLoginRequest);
    connect(networkingAndFiles, &NetworkingAndFiles::showWifiBusy,    this,               &TefCacApplication::networkWifiBusy);

    connect(this, SIGNAL(hideBkg()),                                                 uiMainWindow, SLOT(hideBkg()));
    connect(this, SIGNAL(quranNextAyah(QVariant)),                                   uiMainWindow, SLOT(quranNextAyah(QVariant)));
    connect(this, SIGNAL(setBkg(QVariant)),                                          uiMainWindow, SLOT(setBkg(QVariant)));
    connect(this, SIGNAL(setBkgDirCtrlVisible(QVariant)),                            uiMainWindow, SLOT(settingsBkgDirCtrlVisible(QVariant)));
    connect(this, SIGNAL(setCheckForUpdatesVisible(QVariant)),                       uiMainWindow, SLOT(settingsCheckForUpdatesVisible(QVariant)));
    connect(this, SIGNAL(setCountries(QVariant)),                                    uiMainWindow, SLOT(settingsCountries(QVariant)));
    connect(this, SIGNAL(setCurrentPrayer(QVariant)),                                uiMainWindow, SLOT(setCurrentPrayer(QVariant)));
    connect(this, SIGNAL(setDate(QVariant, QVariant, QVariant)),                     uiMainWindow, SLOT(setDate(QVariant, QVariant, QVariant)));
    connect(this, SIGNAL(setErrorStr(QVariant, QVariant)),                           uiMainWindow, SLOT(setErrorStr(QVariant, QVariant)));
    connect(this, SIGNAL(setFont(QVariant, QVariant, QVariant, QVariant, QVariant)), uiMainWindow, SLOT(setFont(QVariant, QVariant, QVariant, QVariant, QVariant)));
    connect(this, SIGNAL(setLocation(QVariant, QVariant, QVariant, QVariant)),       uiMainWindow, SLOT(setLocation(QVariant, QVariant, QVariant, QVariant)));
    connect(this, SIGNAL(setLocationBusy(QVariant)),                                 uiMainWindow, SLOT(settingsLocationBusy(QVariant)));
    connect(this, SIGNAL(setPrayer(QVariant, QVariant, QVariant, QVariant)),         uiMainWindow, SLOT(setPrayer(QVariant, QVariant, QVariant, QVariant)));
    connect(this, SIGNAL(setQuranAudioToggleRole(QVariant)),                         uiMainWindow, SLOT(quranAudioToggleRole(QVariant)));
    connect(this, SIGNAL(setQuranAyah(QVariant, QVariant)),                          uiMainWindow, SLOT(quranAyah(QVariant, QVariant)));
    connect(this, SIGNAL(setQuranBusy(QVariant)),                                    uiMainWindow, SLOT(quranBusy(QVariant)));
    connect(this, SIGNAL(setQuranButtonVisible(QVariant)),                           uiMainWindow, SLOT(quranButtonVisible(QVariant)));
    connect(this, SIGNAL(setQuranSuwar(QVariant, QVariant, QVariant)),               uiMainWindow, SLOT(quranSuwar(QVariant, QVariant, QVariant)));
    connect(this, SIGNAL(setRegions(QVariant)),                                      uiMainWindow, SLOT(settingsRegions(QVariant)));
    connect(this, SIGNAL(setScreen(QVariant)),                                       uiMainWindow, SLOT(setScreen(QVariant)));
    connect(this, SIGNAL(setSearchResults(QVariant)),                                uiMainWindow, SLOT(settingsSearchResults(QVariant)));
    connect(this, SIGNAL(setUIButtonRole(QVariant)),                                 uiMainWindow, SLOT(setUIButtonRole(QVariant)));
    connect(this, SIGNAL(setTime(QVariant, QVariant, QVariant)),                     uiMainWindow, SLOT(setTime(QVariant, QVariant, QVariant)));
    connect(this, SIGNAL(settingsData(QVariant)),                                    uiMainWindow, SLOT(settingsInit(QVariant)));
    connect(this, SIGNAL(setWifiBusy(QVariant)),                                     uiMainWindow, SLOT(wifiBusy(QVariant)));
    connect(this, SIGNAL(setWifiList(QVariant)),                                     uiMainWindow, SLOT(wifiSetList(QVariant)));

    connect(uiMainWindow, SIGNAL(buttonPressed(int)),                                                         this, SLOT(uiButtonPressed(int)));
    connect(uiMainWindow, SIGNAL(checkForUpdatesPressed()),                                                   this, SLOT(checkUpdatesButtonPressed()));
    connect(uiMainWindow, SIGNAL(quranAudioTogglePressed(int, int, int)),                                     this, SLOT(quranAudioTogglePressed(int, int, int)));
    connect(uiMainWindow, SIGNAL(quranAyahChanged(int, int)),                                                 this, SLOT(quranAyahChanged(int, int)));
    connect(uiMainWindow, SIGNAL(quranPressed()),                                                             this, SLOT(quranButtonPressed()));
    connect(uiMainWindow, SIGNAL(settingsCountrySelected(QString, QString)),                                  this, SLOT(settingsCountrySelected(QString, QString)));
    connect(uiMainWindow, SIGNAL(settingsLocationSearchPressed(QString, QString, QString, QString, QString)), this, SLOT(settingsLocationSearchPressed(QString, QString, QString, QString, QString)));
    connect(uiMainWindow, SIGNAL(settingsSoundTestPressed(int, int)),                                         this, SLOT(settingsSoundTestPressed(int, int)));
    connect(uiMainWindow, SIGNAL(settingsUpdated(QString)),                                                   this, SLOT(settingsUpdated(QString)));
    connect(uiMainWindow, SIGNAL(shutDownPressed()),                                                          this, SLOT(shutDownPressed()));
    connect(uiMainWindow, SIGNAL(wifiSelected(QString, QString)),                                             this, SLOT(wifiSelected(QString, QString)));
    connect(uiMainWindow, SIGNAL(wifiCanceled()),                                                             this, SLOT(wifiCanceled()));

    networkingAndFilesThread.start();

    QSettings settings;

    int screenWidth  = primaryScreen()->geometry().width();
    int screenHeight = primaryScreen()->geometry().height();

    int w = settings.value("ui/w", (screenWidth > 640) && (screenHeight > 480) ? 480 : 0).toInt();
    int h = settings.value("ui/h", (screenWidth > 640) && (screenHeight > 480) ? 320 : 0).toInt();
    if ((w <= 0) || (h <= 0) || envToBool("TEF_CAC_KIOSK_MODE")) {
        uiMainWindow->showFullScreen();
    }
    else {
        uiMainWindow->setWidth(w);
        uiMainWindow->setHeight(h);
    }

    emit setBkgDirCtrlVisible(!envToBool("TEF_CAC_KIOSK_MODE"));
    emit setCheckForUpdatesVisible(envToBool("TEF_CAC_KIOSK_MODE"));
    emit setUIButtonRole(BUTTON_ROLE_SETTINGS);
    emit setQuranButtonVisible(envToBool("TEF_CAC_KIOSK_MODE"));

    splash = true;
    emit setBkg("qrc:///splash.png");

    int nextTimer = 750;
    if (settings.value("prefs/splash_sound", DEFAULT_SPLASH_SOUND).toBool() && playSound(QFileInfo(QString("%1/bismillah.wav").arg(settings.value("media/sound_dir", DEFAULT_MEDIA_DIR).toString())))) {
        nextTimer = 5000;
    }

    mediaPlayer.setVolume(settings.value("prefs/volume", DEFAULT_VOLUME).toInt());

    if (!settings.childGroups().contains("locationCustom")) {
        if (settings.childGroups().contains("locationAuto")) {
            locationDataExists = true;

            QString country      = settings.value("locationAuto/country").toString();
            QString region       = settings.value("locationAuto/region").toString();
            QString city         = settings.value("locationAuto/city").toString();
            QString timeZoneCode = settings.value("locationAuto/timeZoneCode").toString();

            emit setLocation(city, region, country, timeZoneCode);
        }
        emit networkGet(NetworkingAndFiles::QUERY_LOCATION);
    }
    else {
        locationDataExists = true;

        QString country      = settings.value("locationCustom/country").toString();
        QString region       = settings.value("locationCustom/region").toString();
        QString city         = settings.value("locationCustom/city").toString();
        QString timeZoneCode = settings.value("locationCustom/timeZoneCode").toString();

        emit setLocation(city, region, country, timeZoneCode);
    }

    installTranslation();
    updateScreen(true);

    mainTimer.singleShot(nextTimer, this, &TefCacApplication::mainTimerProc);
}


void TefCacApplication::settingsCountrySelected(QString geonameId, QString countryCode)
{
    Q_UNUSED(countryCode);

    emit setLocationBusy(true);
    emit networkGet(NetworkingAndFiles::QUERY_REGIONS, 0, QVariantMap({{ "geonameId", geonameId }}));
}


QString TefCacApplication::settingsFileName()
{
    QSettings settings;
    return settings.fileName();
}


void TefCacApplication::settingsLocationSearchPressed(QString countryGeonameId, QString countryCode, QString regionGeonameId, QString regionCode, QString searchText)
{
    Q_UNUSED(countryGeonameId);
    Q_UNUSED(regionGeonameId);

    emit setLocationBusy(true);

    QVariantMap params;
    params.insert("country", countryCode);

    QRegularExpression      postalRegexp("^\\d+$");
    QRegularExpressionMatch postalMatch = postalRegexp.match(searchText);
    if (postalMatch.hasMatch()) {
        params.insert("postalcode", searchText);
        emit networkGet(NetworkingAndFiles::LOCATION_SEARCH_POSTALCODE, 0, params);
        return;
    }

    params.insert("name_startsWith", searchText);
    if (regionCode.length()) {
        params.insert("adminCode1", regionCode);
    }
    emit networkGet(NetworkingAndFiles::LOCATION_SEARCH_FREETEXT, 0, params);
}


void TefCacApplication::settingsOpen()
{
    QSettings   settings;
    QVariantMap optionsMap;

    QList<AvailableLocale> availableLanguages;
    QFileInfoList entries = QDir(applicationDirPath()).entryInfoList(QDir::Files | QDir::Readable);
    foreach (QFileInfo entry, entries) {
        if (entry.suffix().compare("qm", Qt::CaseInsensitive) == 0) {
            QString localeAbbr = entry.baseName().remove("tef_cac_", Qt::CaseInsensitive);
            QLocale localeObj(localeAbbr);
            if (localeObj.language() == QLocale::C) {
                continue;
            }
            availableLanguages.append({ localeAbbr, QString("%1 (%2)").arg(defLocale.languageToString(localeObj.language())).arg(localeObj.nativeLanguageName()) });
         }
    }
    std::sort(availableLanguages.begin(), availableLanguages.end(), [](AvailableLocale a, AvailableLocale b) {
        return a.name.compare(b.name, Qt::CaseInsensitive) < 0;
    });
    availableLanguages.prepend({"-", tr("Default")});

    QVariantList languages;
    foreach(AvailableLocale availableLanguage, availableLanguages) {
        languages.append(QVariantMap({{ "abbr", availableLanguage.abbr }, { "name", availableLanguage.name }}));
    }

    optionsMap.insert("language",     settings.value("prefs/language", "-"));
    optionsMap.insert("languages",    languages);
    optionsMap.insert("volume",       settings.value("prefs/volume",       DEFAULT_VOLUME));
    optionsMap.insert("splash_sound", settings.value("prefs/splash_sound", DEFAULT_SPLASH_SOUND).toBool());
    optionsMap.insert("use_bkg",      settings.value("prefs/use_bkg",      DEFAULT_USE_BKG).toBool());
    optionsMap.insert("bkg_dir",      settings.value("media/bkg_dir",      QString("%1/bkg").arg(DEFAULT_MEDIA_DIR)));

    QVariantList locations;
    QVariantMap  location = currentLocation();
    if (!location.isEmpty()) {
        locations.append(location);
    }

    optionsMap.insert("location_custom", settings.childGroups().contains("locationCustom"));
    optionsMap.insert("locations", locations);

    optionsMap.insert("gregorian_date_visible",        settings.value("prefs/gregorian_date_visible",        DEFAULT_DATE_GREGORIAN_VISIBLE).toBool());
    optionsMap.insert("hijri_date_visible",            settings.value("prefs/hijri_date_visible",            DEFAULT_DATE_HIJRI_VISIBLE).toBool());
    optionsMap.insert("shamsi_date_visible",           settings.value("prefs/shamsi_date_visible",           DEFAULT_DATE_SHAMSI_VISIBLE).toBool());
    optionsMap.insert("date_format_full",              settings.value("prefs/date_format_full",              DEFAULT_DATE_FORMAT_FULL).toBool());
    optionsMap.insert("gregorian_date_always_english", settings.value("prefs/gregorian_date_always_english", DEFAULT_DATE_GREGORIAN_ALWAYS_ENGLISH).toBool());
    optionsMap.insert("western_arabic_time_visible",   settings.value("prefs/western_arabic_time_visible",   DEFAULT_TIME_WESTERN_ARABIC_VISIBLE).toBool());
    optionsMap.insert("eastern_arabic_time_visible",   settings.value("prefs/eastern_arabic_time_visible",   DEFAULT_TIME_EASTERN_ARABIC_VISIBLE).toBool());
    optionsMap.insert("farsi_time_visible",            settings.value("prefs/farsi_time_visible",            DEFAULT_TIME_FARSI_VISIBLE).toBool());
    optionsMap.insert("time_format_24",                settings.value("prefs/time_format_24",                DEFAULT_TIME_24).toBool());

    QVariantList prayerNames = {
        QVariantMap({{ "id", PRAYER_NAMES_LOCAL },           { "name", tr("Default Language") }}),
        QVariantMap({{ "id", PRAYER_NAMES_ARABIC },          { "name", tr("Arabic") }}),
        QVariantMap({{ "id", PRAYER_NAMES_TRANSLITERATION }, { "name", tr("Transliteration") }})
    };

    optionsMap.insert("prayer_time_names_list",         prayerNames);
    optionsMap.insert("prayer_times_names",             settings.value("prefs/prayer_times_names",             DEFAULT_PRAYER_NAMES).toInt());
    optionsMap.insert("prayer_times_subh",              settings.value("prefs/prayer_times_subh",              DEFAULT_PRAYER_NAMES_SUBH).toBool());
    optionsMap.insert("last_prayer_highlight_minutes",  settings.value("prefs/last_prayer_highlight_minutes",  DEFAULT_HIGHLIGHT_LAST));
    optionsMap.insert("next_prayer_highlight_minutes",  settings.value("prefs/next_prayer_highlight_minutes",  DEFAULT_HIGHLIGHT_NEXT));
    optionsMap.insert("next_shurooq_highlight_minutes", settings.value("prefs/next_shurooq_highlight_minutes", DEFAULT_HIGHLIGHT_SHUROOQ));

    QVariantList prayerSounds = {
        QVariantMap({{ "id", QString("%1").arg(ADHAN) },  { "name", tr("Adhan") }}),
        QVariantMap({{ "id", QString("%1").arg(BEEP) },   { "name", tr("Beep") }}),
        QVariantMap({{ "id", QString("%1").arg(SILENT) }, { "name", tr("Silent") }})
    };
    QVariantList shurooqSounds = {
        QVariantMap({{ "id", QString("%1").arg(SHUROOQ_BEEP) },   { "name", tr("Beep") }}),
        QVariantMap({{ "id", QString("%1").arg(SHUROOQ_SILENT) }, { "name", tr("Silent") }})
    };

    optionsMap.insert("prayer_sounds",  prayerSounds);
    optionsMap.insert("shurooq_sounds", shurooqSounds);
    optionsMap.insert("fajr",           prayerName(FAJR,    true));
    optionsMap.insert("shurooq",        prayerName(SHUROOQ, true));
    optionsMap.insert("dhuhr",          prayerName(DHUHR,   true));
    optionsMap.insert("asr",            prayerName(ASR,     true));
    optionsMap.insert("maghrib",        prayerName(MAGHRIB, true));
    optionsMap.insert("isha",           prayerName(ISHA,    true));
    optionsMap.insert("sound_fajr",     QString("%1").arg(settings.value("prefs/sound_fajr",    DEFAULT_PRAYER_SOUND).toInt()));
    optionsMap.insert("sound_shurooq",  QString("%1").arg(settings.value("prefs/sound_shurooq", DEFAULT_SHUROOQ_SOUND).toInt()));
    optionsMap.insert("sound_dhuhr",    QString("%1").arg(settings.value("prefs/sound_dhuhr",   DEFAULT_PRAYER_SOUND).toInt()));
    optionsMap.insert("sound_asr",      QString("%1").arg(settings.value("prefs/sound_asr",     DEFAULT_PRAYER_SOUND).toInt()));
    optionsMap.insert("sound_maghrib",  QString("%1").arg(settings.value("prefs/sound_maghrib", DEFAULT_PRAYER_SOUND).toInt()));
    optionsMap.insert("sound_isha",     QString("%1").arg(settings.value("prefs/sound_isha",    DEFAULT_PRAYER_SOUND).toInt()));

    optionsMap.insert("prayer_times_method",           QString("%1").arg(settings.value("prefs/prayer_times_method",           DEFAULT_PRAYER_METHOD).toInt()));
    optionsMap.insert("prayer_times_school",           QString("%1").arg(settings.value("prefs/prayer_times_school",           DEFAULT_PRAYER_SCHOOL).toInt()));
    optionsMap.insert("prayer_times_high_latitude",    QString("%1").arg(settings.value("prefs/prayer_times_high_latitude",    DEFAULT_HIGH_LATITUDE).toInt()));
    optionsMap.insert("prayer_times_extreme_latitude", QString("%1").arg(settings.value("prefs/prayer_times_extreme_latitude", DEFAULT_EXTREME_LATITUDE).toInt()));

    QVariantList calculationMethods = {
        QVariantMap({{ "id", QString("%1").arg(METHOD_EGYPT) },             { "name", tr("Al-Azhar University (Egypt)") }}),
        QVariantMap({{ "id", QString("%1").arg(METHOD_IRAN) },              { "name", tr("Leva Research Institude (Iran)") }}),
        QVariantMap({{ "id", QString("%1").arg(METHOD_NORTH_AMERICA) },     { "name", tr("Islamic Society of North America (ISNA)") }}),
        QVariantMap({{ "id", QString("%1").arg(METHOD_PAKISTAN) },          { "name", tr("University Of Karachi (Pakistan)") }}),
        QVariantMap({{ "id", QString("%1").arg(METHOD_REST_OF_THE_WORLD) }, { "name", tr("Muslim World League (MWL)") }}),
        QVariantMap({{ "id", QString("%1").arg(METHOD_SAUDI) },             { "name", tr("Umm Al-Qura (Saudi Arabia)") }})
    };
    std::sort(calculationMethods.begin(), calculationMethods.end(), [](QVariant a, QVariant b) {
        return a.toMap().value("name").toString().compare(b.toMap().value("name").toString(), Qt::CaseInsensitive) < 0;
    });
    calculationMethods.prepend(QVariantMap({{ "id", QString("%1").arg(METHOD_AUTODETECT) },  { "name", tr("Detect Automatically") }}));

    QVariantList calculationSchools = {
        QVariantMap({{ "id", QString("%1").arg(SCHOOL_HANAFI) },               { "name", tr("Hanafi") }}),
        QVariantMap({{ "id", QString("%1").arg(SCHOOL_HANBALI_MALIKI_SHAFI) }, { "name", tr("Hanbali, Maliki, Shafi") }}),
    };
    std::sort(calculationSchools.begin(), calculationSchools.end(), [](QVariant a, QVariant b) {
        return a.toMap().value("name").toString().compare(b.toMap().value("name").toString(), Qt::CaseInsensitive) < 0;
    });
    calculationSchools.prepend(QVariantMap({{ "id", QString("%1").arg(SCHOOL_AUTODETECT) },  { "name", tr("Detect Automatically") }}));

    QVariantList highLatitudes = {
        QVariantMap({{ "id", QString("%1").arg(HIGH_LATITUDE_NONE) },              { "name", tr("None") }}),
        QVariantMap({{ "id", QString("%1").arg(HIGH_LATITUDE_MIDDLE_OF_NIGHT) },   { "name", tr("Middle of the night") }}),
        QVariantMap({{ "id", QString("%1").arg(HIGH_LATITUDE_SEVENTH_OF_NIGHT) },  { "name", tr("One-seventh of the night") }}),
        QVariantMap({{ "id", QString("%1").arg(HIGH_LATITUDE_FRACTION_OF_NIGHT) }, { "name", tr("Fraction of the night (depending on method)") }}),
    };
    std::sort(highLatitudes.begin(), highLatitudes.end(), [](QVariant a, QVariant b) {
        return a.toMap().value("name").toString().compare(b.toMap().value("name").toString(), Qt::CaseInsensitive) < 0;
    });
    highLatitudes.prepend(QVariantMap({{ "id", QString("%1").arg(HIGH_LATITUDE_DEFAULT) },  { "name", tr("Default") }}));

    QVariantList extremeLatitudes = {
        QVariantMap({{ "id", QString("%1").arg(EXTREME_LATITUDE_NONE) },             { "name", tr("None") }}),
        QVariantMap({{ "id", QString("%1").arg(EXTREME_LATITUDE_NEAREST_DAY) },      { "name", tr("Nearest Day") }}),
        QVariantMap({{ "id", QString("%1").arg(EXTREME_LATITUDE_NEAREST_LATITUDE) }, { "name", tr("Nearest Latitude") }}),
    };
    std::sort(extremeLatitudes.begin(), extremeLatitudes.end(), [](QVariant a, QVariant b) {
        return a.toMap().value("name").toString().compare(b.toMap().value("name").toString(), Qt::CaseInsensitive) < 0;
    });
    extremeLatitudes.prepend(QVariantMap({{ "id", QString("%1").arg(EXTREME_LATITUDE_DEFAULT) },  { "name", tr("Default") }}));

    optionsMap.insert("calculation_methods", calculationMethods);
    optionsMap.insert("calculation_schools", calculationSchools);
    optionsMap.insert("high_latitudes",      highLatitudes);
    optionsMap.insert("extreme_latitudes",   extremeLatitudes);

    optionsMap.insert("version", QString("%1 %2").arg(tr("Version")).arg(applicationVersion()));

    emit settingsData(optionsMap);
    emit setScreen(SCREEN_SETTINGS);   // TODO pause main screen and background updates (but not adhan sound)

    if (!countriesQueried) {
        emit networkGet(NetworkingAndFiles::QUERY_COUNTRIES);
    }
}


void TefCacApplication::settingsUpdated(QString settingsJSON)
{
    QSettings settings;
    QVariantMap newSettings = QJsonDocument::fromJson(settingsJSON.toUtf8()).toVariant().toMap();

    bool languageChanged = newSettings.value("language").toString().compare(settings.value("prefs/language", "-").toString(), Qt::CaseInsensitive);

    settings.setValue("prefs/language",     newSettings.value("language").toString());
    settings.setValue("prefs/volume",       newSettings.value("volume").toInt());
    settings.setValue("prefs/splash_sound", newSettings.value("splash_sound").toBool());
    settings.setValue("prefs/use_bkg",      newSettings.value("use_bkg").toBool());
    settings.setValue("media/bkg_dir",      newSettings.value("bkg_dir").toString());

    settings.remove("locationAuto");
    settings.remove("locationCustom");

    settings.setValue("prefs/gregorian_date_visible",        newSettings.value("gregorian_date_visible").toBool());
    settings.setValue("prefs/hijri_date_visible",            newSettings.value("hijri_date_visible").toBool());
    settings.setValue("prefs/shamsi_date_visible",           newSettings.value("shamsi_date_visible").toBool());
    settings.setValue("prefs/date_format_full/",             newSettings.value("date_format_full").toBool());
    settings.setValue("prefs/gregorian_date_always_english", newSettings.value("gregorian_date_always_english").toBool());
    settings.setValue("prefs/western_arabic_time_visible",   newSettings.value("western_arabic_time_visible").toBool());
    settings.setValue("prefs/eastern_arabic_time_visible",   newSettings.value("eastern_arabic_time_visible").toBool());
    settings.setValue("prefs/farsi_time_visible",            newSettings.value("farsi_time_visible").toBool());
    settings.setValue("prefs/time_format_24",                newSettings.value("time_format_24").toBool());

    settings.setValue("prefs/prayer_times_names",             newSettings.value("prayer_times_names").toInt());
    settings.setValue("prefs/prayer_times_subh",              newSettings.value("prayer_times_subh").toBool());
    settings.setValue("prefs/last_prayer_highlight_minutes",  newSettings.value("last_prayer_highlight_minutes").toInt());
    settings.setValue("prefs/next_prayer_highlight_minutes",  newSettings.value("next_prayer_highlight_minutes").toInt());
    settings.setValue("prefs/next_shurooq_highlight_minutes", newSettings.value("next_shurooq_highlight_minutes").toInt());

    settings.setValue("prefs/sound_fajr",    newSettings.value("sound_fajr").toInt());
    settings.setValue("prefs/sound_shurooq", newSettings.value("sound_shurooq").toInt());
    settings.setValue("prefs/sound_dhuhr",   newSettings.value("sound_dhuhr").toInt());
    settings.setValue("prefs/sound_asr",     newSettings.value("sound_asr").toInt());
    settings.setValue("prefs/sound_maghrib", newSettings.value("sound_maghrib").toInt());
    settings.setValue("prefs/sound_isha",    newSettings.value("sound_isha").toInt());

    settings.setValue("prefs/prayer_times_method",           newSettings.value("prayer_times_method").toInt());
    settings.setValue("prefs/prayer_times_school",           newSettings.value("prayer_times_school").toInt());
    settings.setValue("prefs/prayer_times_high_latitude",    newSettings.value("prayer_times_high_latitude").toInt());
    settings.setValue("prefs/prayer_times_extreme_latitude", newSettings.value("prayer_times_extreme_latitude").toInt());

    settings.sync();

    if (languageChanged) {
        installTranslation();
        suwarInitialized = false;
    }
    mediaPlayer.setVolume(newSettings.value("volume").toInt());

    if (newSettings.value("location_custom").toBool()) {
        QString tzCode = newSettings.value("location_selected", {}).toMap().value("timeZoneCode", "").toString();
        if (tzCode.isEmpty()) {
            QVariantMap params;
            params.insert("lat",         newSettings.value("location_selected", {}).toMap().value("latitude"));
            params.insert("lng",         newSettings.value("location_selected", {}).toMap().value("longitude"));
            params.insert("country",     newSettings.value("location_selected", {}).toMap().value("country"));
            params.insert("countryCode", newSettings.value("location_selected", {}).toMap().value("countryCode"));
            params.insert("region",      newSettings.value("location_selected", {}).toMap().value("region"));
            params.insert("city",        newSettings.value("location_selected", {}).toMap().value("city"));

            emit networkGet(NetworkingAndFiles::LOCATION_TIMEZONE, 0, params);
        }
        else {
            QString country      = newSettings.value("location_selected", {}).toMap().value("country").toString();
            QString region       = newSettings.value("location_selected", {}).toMap().value("region").toString();
            QString city         = newSettings.value("location_selected", {}).toMap().value("city").toString();
            QString timeZone     = newSettings.value("location_selected", {}).toMap().value("timeZone").toString();

            settings.setValue("locationCustom/latitude",     newSettings.value("location_selected", {}).toMap().value("latitude").toDouble());
            settings.setValue("locationCustom/longitude",    newSettings.value("location_selected", {}).toMap().value("longitude").toDouble());
            settings.setValue("locationCustom/country",      country);
            settings.setValue("locationCustom/countryCode",  newSettings.value("location_selected", {}).toMap().value("countryCode").toString());
            settings.setValue("locationCustom/region",       region);
            settings.setValue("locationCustom/city",         city);
            settings.setValue("locationCustom/timeZone",     timeZone);
            settings.setValue("locationCustom/timeZoneCode", tzCode);

            settings.sync();

            timeZoneCache = QTimeZone(timeZone.toUtf8());

            emit setLocation(city, region, country, tzCode);
        }
    }
    else {
        emit networkGet(NetworkingAndFiles::QUERY_LOCATION);
    }

    if (settings.value("prefs/use_bkg", DEFAULT_USE_BKG).toBool()) {
        backgroundIndex = 0;
        swapBackground();
        backgroundTimer.start(BACKGROUND_DELAY_MS);
    }
    else {
        backgroundTimer.stop();
        emit hideBkg();
    }

    updateScreen(true);
}


void TefCacApplication::settingsSoundTestPressed(int prayer, int sound)
{
    playAdhanSound(static_cast<PRAYER_EVENTS>(prayer), sound);
}

void TefCacApplication::swapBackground()
{
    QSettings settings;

    QDir dir(settings.value("media/bkg_dir", QString("%1/bkg").arg(DEFAULT_MEDIA_DIR)).toString());
    if (!dir.exists()) {
        emit hideBkg();
        backgroundTimer.stop();
        return;
    }

    QFileInfoList images;
    QFileInfoList entries = dir.entryInfoList(QDir::Files | QDir::Readable);
    foreach (QFileInfo entry, entries) {
        if (mimeDatabase.mimeTypeForFile(entry).name().startsWith("image")) {
            images.append(entry);
        }
    }

    if (images.isEmpty()) {
        emit hideBkg();
        backgroundTimer.stop();
        return;
    }

    std::sort(images.begin(), images.end(), [](QFileInfo a, QFileInfo b) {
        return a.baseName().compare(b.baseName(), Qt::CaseInsensitive) < 0;
    });

    if (backgroundIndex >= images.size()) {
        backgroundIndex = 0;
    }
    emit setBkg(QUrl::fromLocalFile(images.at(backgroundIndex).absoluteFilePath()));
    backgroundIndex++;
}


void TefCacApplication::shutDownPressed()
{
    exit(EXIT_CODE_SHUT_DOWN);
}


void TefCacApplication::uiButtonPressed(int role)
{
    if (role == BUTTON_ROLE_SETTINGS) {
        settingsOpen();
    }
    else if (role == BUTTON_ROLE_STOP) {
        mediaPlayer.stop();
    }
}


void TefCacApplication::updatePrayerTimes(bool forcedRecalculate, PrayerTimes::PrayerEvent next)
{
    // handleErrorMsg("", false);

    bool updateAllPrayers = false;

    if (forcedRecalculate || !prayerTimes.isValid()) {
        prayerTimes.calculate();
        updateAllPrayers = true;
    }
    if (!prayerTimes.isValid()) {
        QString errMsg = prayerTimes.lastError();
        handleErrorMsg(errMsg.isEmpty() ? "Unable to calculate prayer times" : errMsg, false);
        return;
    }
    if (updateAllPrayers || next.time.isNull()) {
        next = prayerTimes.getNext();
    }

    QSettings settings;

    QHash<PRAYER_EVENTS, int> eventsOrder;
    if ((settings.value("prefs/prayer_times_names", DEFAULT_PRAYER_NAMES).toInt() == PRAYER_NAMES_ARABIC) || (defLocale.script() == QLocale::ArabicScript)) {
        eventsOrder = {
            { FAJR,    6 },
            { SHUROOQ, 5 },
            { DHUHR,   4 },
            { ASR,     3 },
            { MAGHRIB, 2 },
            { ISHA,    1 }
        };
    }
    else {
        eventsOrder = {
            { FAJR,    1 },
            { SHUROOQ, 2 },
            { DHUHR,   3 },
            { ASR,     4 },
            { MAGHRIB, 5 },
            { ISHA,    6 }
        };
    }

    PrayerTimes::PrayerEvent previous = { NONE, QTime(), 0, 0 };

    if (!updateAllPrayers) {
        previous = prayerTimes.getPrevious();
        if ((next.percent > 0) && (previous.percent == 0)) {
            emit setPrayer(eventsOrder.value(next.prayerEvent), prayerName(next.prayerEvent), prayerTimeFormatted(next.time, true), 1 - next.percent);
            return;
        }
    }

    if (previous.time.isNull()) {
        previous = prayerTimes.getPrevious();
    }
    if (previous.percent > 0) {
        if (!soundPlayed) {
            emit setCurrentPrayer(prayerName(previous.prayerEvent, true));
            if (previous.percent <= 0.25) {
                playAdhan(previous.prayerEvent);
            }
            soundPlayed = true;
        }
        return;
    }
    if (soundPlayed) {
        soundPlayed      = false;
        updateAllPrayers = true;
    }
    if (!updateAllPrayers) {
        return;
    }

    foreach (PRAYER_EVENTS event, QList<PRAYER_EVENTS>({ FAJR, SHUROOQ, DHUHR, ASR, MAGHRIB, ISHA })) {
        if ((event == next.prayerEvent) && (next.percent > 0)) {
            emit setPrayer(eventsOrder.value(next.prayerEvent), prayerName(next.prayerEvent), prayerTimeFormatted(next.time, true), 1 - next.percent);
            continue;
        }

        QTime  time = prayerTimes.getPrayer(event);
        emit setPrayer(eventsOrder.value(event), prayerName(event), prayerTimeFormatted(time), 0);
    }
}


void TefCacApplication::updateScreen(bool forced)
{
    QTime nowTime = getCurrentDateTime().time();

    if ((nowTime.hour() == 2) && (nowTime.minute() == 0) && (hijriProcess.state() == QProcess::NotRunning) && (envToBool("TEF_CAC_KIOSK_MODE"))) {
        // TODO save "last checked" to settings
        exit(EXIT_CODE_CHECK_UPDATES);
        return;
    }

    if (!forced && screenLastUpdated.isValid() && (nowTime.second() == screenLastUpdated.second())) {
        return;
    }

    PrayerTimes::PrayerEvent next = prayerTimes.getNext();

    if (!forced && screenLastUpdated.isValid() && (nowTime.minute() == screenLastUpdated.minute()) && (next.percent == 0)) {
        return;
    }

    QSettings settings;

    if (forced || !screenLastUpdated.isValid() || (nowTime.minute() != screenLastUpdated.minute())) {
        bool time24 = settings.value("prefs/time_format_24", false).toBool();
        int  hours  = nowTime.hour();
        if (!time24) {
            if (hours > 12) {
                hours -= 12;
            }
            if (hours == 0) {
                hours = 12;
            }
        }
        QString timeStr = QString("%1:%2").arg(hours, time24 ? 2 : 0, 10, QChar('0')).arg(nowTime.minute(), 2, 10, QChar('0'));

        QString western;
        QString eastern;
        QString farsi;
        if (settings.value("prefs/western_arabic_time_visible", DEFAULT_TIME_WESTERN_ARABIC_VISIBLE).toBool()) {
            western = timeStr;
        }
        if (settings.value("prefs/eastern_arabic_time_visible", DEFAULT_TIME_EASTERN_ARABIC_VISIBLE).toBool()) {
            eastern = Globals::convertNumbersToEastern(timeStr);
        }
        if (settings.value("prefs/farsi_time_visible", DEFAULT_TIME_FARSI_VISIBLE).toBool()) {
            farsi = Globals::convertNumbersToFarsi(timeStr);
        }
        emit setTime(western, eastern, farsi);
    }

    if (screenLastUpdated.isValid() && !forced && !((nowTime.hour() == 0) && (nowTime.minute() == 0))) {
        if ((next.percent != 0) || (nowTime.minute() != screenLastUpdated.minute())) {
            screenLastUpdated.setHMS(nowTime.hour(), nowTime.minute(), nowTime.second());
            updatePrayerTimes(false, next);
        }
        return;
    }

    screenLastUpdated.setHMS(nowTime.hour(), nowTime.second(), nowTime.minute());

    if (hijriProcess.state() == QProcess::NotRunning) {
        hijriProcess.start(settings.value("ext/hijri_conv", DEFAULT_HIJRI_CONVERTER).toString());
    }
    updatePrayerTimes(true);
}


void TefCacApplication::wifiSelected(QString ssid, QString psw)
{
    emit wifiConnect(ssid, psw);
}


void TefCacApplication::wifiCanceled()
{
    emit wifiConnect("", "");
}
