#ifndef TEFCACAPPLICATION_H
#define TEFCACAPPLICATION_H

#include <QDate>
#include <QDateTime>
#include <QFontDatabase>
#include <QGuiApplication>
#include <QIcon>
#include <QList>
#include <QLocale>
#include <QMediaPlayer>
#include <QMimeDatabase>
#include <QObject>
#include <QQmlApplicationEngine>
#include <QQuickWindow>
#include <QProcess>
#include <QScreen>
#include <QSettings>
#include <QString>
#include <QStringList>
#include <QThread>
#include <QTime>
#include <QTimer>
#include <QTimeZone>
#include <QTranslator>
#include <QVariant>
#include <QVariantList>
#include <QVariantMap>

#include "globals.h"
#include "networkingandfiles.h"
#include "prayertimes.h"


class TefCacApplication : public QGuiApplication
{
    Q_OBJECT

    public:

        explicit TefCacApplication(int &argc, char **argv);
        ~TefCacApplication();

        void setQmlApplicationEngine(QQmlApplicationEngine *qmlApplicationEngine);

        QString settingsFileName();


    private:

        struct AvailableLocale {
            QString abbr;
            QString name;
        };

        static const int BUTTON_ROLE_SETTINGS = 1;
        static const int BUTTON_ROLE_STOP     = 2;
        static const int BUTTON_ROLE_PLAY     = 3;

        static const int EXIT_CODE_CHECK_UPDATES = 255;
        static const int EXIT_CODE_SHUT_DOWN     = 254;

        static const int VERSION_MAJOR = 0;
        static const int VERSION_MINOR = 4;
        static const int VERSION_SUB   = 4;

        static const int SCREEN_MAIN     = 0;
        static const int SCREEN_SETTINGS = 1;
        static const int SCREEN_QURAN    = 2;
        static const int SCREEN_WIFI     = 3;

        static const int BACKGROUND_DELAY_MS   = 12 * 60 * 1000;
        static const int MAX_LOADED_FONTS      = 25;

        QQmlApplicationEngine *qmlApplicationEngine;
        QMimeDatabase          mimeDatabase;

        QTimer    mainTimer;
        bool      splash;
        int       screenAfterSplash;
        bool      countriesQueried;
        bool      locationDataExists;
        QTime     screenLastUpdated;
        bool      soundPlayed;
        QTimeZone timeZoneCache;

        QTimer backgroundTimer;
        int    backgroundIndex;

        QMediaPlayer mediaPlayer;
        int          adhanIndex;
        int          adhanFajrIndex;
        bool         playStartedByQuran;

        QThread             networkingAndFilesThread;
        NetworkingAndFiles *networkingAndFiles;

        QProcess     hijriProcess;
        PrayerTimes  prayerTimes;
        QLocale      defLocale;
        QTranslator* translator;

        bool           suwarInitialized;
        int            ayatMemCacheSurah;
        QVariantList   ayatMemCache;
        QMap<int, int> fontsLoaded;

        void settingsOpen();

        void installTranslation();
        void updateScreen(bool forced = false);
        void updatePrayerTimes(bool forcedRecalculate = false, PrayerTimes::PrayerEvent next = { NONE, QTime(), 0, 0 });
        void playAdhan(PRAYER_EVENTS prayer);
        void playAdhanSound(PRAYER_EVENTS prayer, int sound);
        bool playSound(QFileInfo soundFile);

        void handleErrorMsg(QString msg, bool isTemporary);

        QVariantMap currentLocation();
        bool        envToBool(QString envVar);
        QDateTime   getCurrentDateTime();
        QString     getUILanguage();
        QString     prayerName(PRAYER_EVENTS prayer, bool full = false);
        QString     prayerTimeFormatted(QTime prayerTime, bool remainingtime = false);

        QVariantMap quranFindAyahInMemCache(int ayahNumber);
        void        quranSetUIAyah(QVariantMap ayah);


    private slots:

        void mainTimerProc();
        void swapBackground();
        void mediaStateChanged(QMediaPlayer::State state);
        void mediaError(QMediaPlayer::Error error);

        void hijriFinished(int exitCode, QProcess::ExitStatus exitStatus);
        void hijriError(QProcess::ProcessError error);

        void networkGot(NetworkingAndFiles::QUERIES query, bool success, QVariant result);
        void networkUIMsg(QString msg, bool isTemporary);
        void networkWifiLoginRequest(QStringList ssids);
        void networkWifiBusy(bool busy);


    public slots:

        void uiButtonPressed(int role);
        void checkUpdatesButtonPressed();
        void shutDownPressed();
        void quranButtonPressed();
        void quranAyahChanged(int surah, int ayah);
        void quranAudioTogglePressed(int role, int surah, int ayah);
        void settingsCountrySelected(QString geonameId, QString countryCode);
        void settingsLocationSearchPressed(QString countryGeonameId, QString countryCode, QString regionGeonameId, QString regionCode, QString searchText);
        void settingsSoundTestPressed(int prayer, int sound);
        void settingsUpdated(QString settingsJSON);
        void wifiSelected(QString ssid, QString psw);
        void wifiCanceled();


    signals:

        void networkGet(NetworkingAndFiles::QUERIES query, int page = 0, QVariantMap querySpecific = QVariantMap());
        void logErr(QString msg);

        void hideBkg();
        void quranNextAyah(QVariant sendAudioToggle);
        void setBkg(QVariant image);
        void setBkgDirCtrlVisible(QVariant isVisible);
        void setCheckForUpdatesVisible(QVariant isVisible);
        void setCountries(QVariant countires);
        void setCurrentPrayer(QVariant prayerName);
        void setDate(QVariant gregorian, QVariant hijri, QVariant shamsi);
        void setErrorStr(QVariant errorStr, QVariant isTemporary);
        void setFont(QVariant latinFamily, QVariant arabicFamily, QVariant size, QVariant prayersArabic, QVariant settingsArabic);
        void setLocation(QVariant city, QVariant region, QVariant country, QVariant timezone);
        void setLocationBusy(QVariant busy);
        void setPrayer(QVariant number, QVariant name, QVariant timeString, QVariant highlightPercent);
        void setQuranAudioToggleRole(QVariant isStop);
        void setQuranBusy(QVariant busy);
        void setQuranButtonVisible(QVariant visible);
        void setQuranSuwar(QVariant suwar, QVariant selected, QVariant ayahSelected);
        void setQuranAyah(QVariant text, QVariant fontFamils);
        void setRegions(QVariant countires);
        void setScreen(QVariant screen);
        void setSearchResults(QVariant results);
        void setUIButtonRole(QVariant uiRole);
        void setTime(QVariant westernTime, QVariant easternTime, QVariant farsiTime);
        void settingsData(QVariant optionsMap);
        void wifiConnect(QString ssid, QString psw);
        void setWifiList(QVariant ssidsList);
        void setWifiBusy(QVariant busy);
};

#endif // TEFCACAPPLICATION_H
